from typing import Optional
from scanpy._compat import Literal
from anndata import AnnData
from scanpy import logging as logg

from statsmodels.stats.multitest import multipletests
import multiprocessing as mp
import scipy.sparse
import scipy.stats
import numpy as np

# this whole implementation is modelled after the original R code
# https://github.com/willtownes/scrna2019/blob/master/util/functions.R


def poisson_chunk(inds, Y, n):
    """
    Compute Poisson deviance for a chunk of the count matrix.

    Returns a vector with the per-gene deviance of the chunk.

    Input:
    ------
            inds - a pair of integer indices, defining the start and end gene indices for the chunk
            Y - count matrix, raw counts, cells as rows, genes as columns
            n - what appears as sz in the R implementation, for deviance computation
    """
    # what appears here as n appears as sz in the original R implementation
    # and mu got renamed to lambda
    # subset input data to specified genes for the chunk
    Yn = Y[:, inds[0] : inds[1]]
    # turn to dense if sparse. set up appropriate dimensionality for mu
    if scipy.sparse.issparse(Yn):
        Yn = np.asarray(Yn.todense())
    mu = np.sum(Yn, axis=0) / np.sum(n)
    mu = mu[None, :]
    # in R, they skip all the nan values from the thing defined here as holder
    # seeing how it gets multiplied by the values from Y, instead of trying to find some weird skip
    # just set all the zero values at this step to 1. this way they become 0 after logging
    # and successfully go away in the multiplication. gives same results as matching R command
    holder = Yn / (n * mu)
    holder[holder == 0] = 1
    deviance = 2 * np.sum(Yn * np.log(holder), axis=0) - 2 * np.sum(Yn - n * mu, axis=0)
    return deviance


def binomial_chunk(inds, Y, n):
    """
    Compute binomial deviance for a chunk of the count matrix.

    Returns a vector with the per-gene deviance of the chunk.

    Input:
    ------
            inds - a pair of integer indices, defining the start and end gene indices for the chunk
            Y - count matrix, raw counts, cells as rows, genes as columns
            n - n in the R implementation, for deviance computation
    """
    # subset input data to specified genes for the chunk
    Yn = Y[:, inds[0] : inds[1]]
    # turn to dense if sparse. set up appropriate dimensionality for p
    if scipy.sparse.issparse(Yn):
        Yn = np.asarray(Yn.todense())
    p = np.sum(Yn, axis=0) / np.sum(n)
    p = p[None, :]
    # in R, they skip all the nan values from the thing defined here as holder
    # seeing how it gets multiplied by the values from Y, instead of trying to find some weird skip
    # just set all the zero values at this step to 1. this way they become 0 after logging
    # and successfully go away in the multiplication. gives same results as matching R command
    holder = Yn / (n * p)
    holder[holder == 0] = 1
    term1 = np.sum(Yn * np.log(holder), axis=0)
    nx = n - Yn
    # same thing, second holder
    holder = nx / (n * (1 - p))
    holder[holder == 0] = 1
    term2 = np.sum(nx * np.log(holder), axis=0)
    return 2 * (term1 + term2)


# initiating global variables for the chunk processes to use
def _pool_init(_Y, _n, _chunk_func):
    global Y, n, chunk_func
    Y = _Y
    n = _n
    chunk_func = _chunk_func


# a function that calls the requisite chunk computation based on inds and global variables
def pool_func(inds):
    return chunk_func(inds, Y, n)


def compute_deviance(Y, flavor='poisson', chunksize=1e8, n_jobs=None):
    """
    A function to compute the Poisson/binomial deviance of a count matrix.

    Returns a vector of one deviance per gene

    Input:
    ------
            Y : ``scipy.sparse`` array  or ``numpy.array``
                    Observations as rows, features as columns, raw count (integer) data
            all other input as in highly_deviant_genes()
    """
    # if no info on n_jobs, just go for all the cores
    if n_jobs is None:
        n_jobs = mp.cpu_count()
    # computation prep - obtain n(bin)/sz(poiss), store in variable called n regardless
    # and point chunk_func to appropriate function that computes the deviance of chunks
    if flavor == 'poisson':
        chunk_func = poisson_chunk
        # this n was sz in the R version
        n = np.log(np.sum(Y, axis=1))
        n = np.exp(n - np.mean(n))
    elif flavor == 'binomial':
        chunk_func = binomial_chunk
        n = np.sum(Y, axis=1)
    else:
        raise ValueError('incorrect flavor value')
    # ensure that n has the correct dimensionality, needs to be treated differently if sparse counts
    if scipy.sparse.issparse(Y):
        n = np.asarray(n)
    else:
        n = n[:, None]
    # how many genes per job? we're aiming for chunksize total values in memory at a time
    # so divide that by n_jobs and the number of cells to get genes per job
    chunkcount = np.ceil(chunksize / (Y.shape[0] * n_jobs))
    # set up chunkcount-wide index intervals for count matrix subsetting within the jobs
    inds = []
    for ind in np.arange(0, Y.shape[1], chunkcount):
        inds.append([np.int(ind), np.int(ind + chunkcount)])
    # set up the parallel computation pool with the count matrix and n
    # chunk_func is set to the corresponding Poisson/binomial computation function
    p = mp.Pool(n_jobs, _pool_init, (Y, n, chunk_func))
    deviance = p.map(pool_func, inds)
    # collapse output and return
    return np.hstack(deviance)


def highly_deviant_genes(
    adata: AnnData,
    layer: Optional[str] = None,
    flavor: Literal['poisson', 'binomial'] = 'poisson',
    alpha: Optional[float] = 0.05,
    n_top_genes: Optional[int] = None,
    chunksize: Optional[int] = 1e8,
    n_jobs: Optional[int] = None,
):
    """
    Computation of highly deviant genes [Townes19]_

    Expects count data with unexpressed genes removed.

    Identifies genes statistically significant in a Chi-square test with the number of
    cells informing the number of degrees of freedom (FDR correction via Benjamini-Hochberg).
    If ``n_top_genes`` is provided, instead reports the user-specified number of genes with
    the highest deviance.

    Parameters
    ----------
    adata
            Object to compute highly deviant genes for. Requires raw count space with
            unexpressed genes removed.
    layer
            If provided, use ``adata.layers[layer]`` for expression values instead of ``adata.X``.
    flavor
            Choose between "poisson" and "binomial" deviance
    alpha
            Significance threshold for Benjamini-Hochberg corrected Chi-square testing of
            the deviances, with the number of cells informing the number of degrees of
            freedom. Ignored if ``n_top_genes`` is provided.
    n_top_genes
            If provided, skip the statistical testing and report these many highest
            deviances as the highly deviant genes instead.
    chunksize
            How many values from the count matrix to process at once (across all parallel
            processes). Higher values mean fewer total matrix operations necessary, but
            constrained by memory availability.
    n_jobs
            How many parallel jobs to use for the parallelisation. If ``None``, use all
            available cores.

    Returns
    -------
    Updates ``.var`` with the following fields.

    highly_deviant : bool
            A Boolean indicator of highly deviant genes.
    deviance : float
            The calculated deviance for each of the genes.
    deviance_pval : float
            The Chi-square p-value of the gene being highly deviant; only computed if
            ``n_top_genes`` is ``None``.
    deviance_qval : float
            The p-value after multiple testing correction via Benjamini-Hochberg, used to
            identify the highly deviant genes; only computed if ``n_top_genes`` is ``None``.
    """
    # logging
    start = logg.info('extracting highly deviant genes')
    # run the deviance computation. store in object
    if layer is not None:
        deviance = compute_deviance(
            adata.layers[layer], flavor=flavor, chunksize=chunksize, n_jobs=n_jobs
        )
    else:
        deviance = compute_deviance(
            adata.X, flavor=flavor, chunksize=chunksize, n_jobs=n_jobs
        )
    adata.var['deviance'] = deviance
    adata.uns['hdg'] = {'flavor': flavor}
    # are we just picking the top N genes?
    if n_top_genes is not None:
        # create a vector with highly variable information to insert
        highly_deviant = np.array([False] * adata.shape[1])
        # get the indices of the elements sorted in descending order, then get the top ones
        highly_deviant[np.argsort(deviance)[::-1][:n_top_genes]] = True
        adata.var['highly_deviant'] = highly_deviant
        adata.uns['hdg']['n_top_genes'] = n_top_genes
        # logging
        logg.info('    finished', time=start)
        logg.hint(
            'added\n'
            '    \'highly_deviant\', boolean vector (adata.var)\n'
            '    \'deviance\', float vector (adata.var)\n'
        )
    else:
        # statistical testing - run a chi-square test, then get a benjamini-hochberg correction
        pvals = 1 - scipy.stats.chi2.cdf(deviance, adata.shape[1] - 1)
        qvals = multipletests(pvals, method="fdr_bh")[1]
        adata.var['deviance_pval'] = pvals
        adata.var['deviance_qval'] = qvals
        adata.var['highly_deviant'] = qvals <= alpha
        adata.uns['hdg']['alpha'] = alpha
        # logging
        logg.info('    finished', time=start)
        logg.hint(
            'added\n'
            '    \'highly_deviant\', boolean vector (adata.var)\n'
            '    \'deviance\', float vector (adata.var)\n'
            '    \'deviance_pval\', float vector (adata.var)\n'
            '    \'deviance_qval\', float vector (adata.var)\n'
        )


"""
scanpy filter
"""

import logging
import re
import numpy as np
import scanpy as sc


def filter_anndata(
    adata,
    gene_name="index",
    list_attr=False,
    param=None,
    category=None,
    subset=None,
    force_recalc=False,
):
    """
    Wrapper function for sc.pp.filter_cells() and sc.pp.filter_genes(), mainly
    for supporting arbitrary filtering
    """
    param = [] if param is None else param
    category = [] if category is None else category
    subset = [] if subset is None else subset

    logging.debug("--gene-name=%s", gene_name)
    logging.debug("--param=%s", param)
    logging.debug("--category=%s", category)
    logging.debug("--subset=%s", subset)

    if "mito" not in adata.var.keys() and gene_name:
        try:
            gene_names = getattr(adata.var, gene_name)
            k_mito = gene_names.str.startswith("MT-")
            if k_mito.sum() > 0:
                adata.var["mito"] = k_mito
                adata.var["mito"] = adata.var["mito"].astype("category")
            else:
                logging.warning(
                    "No MT genes found, skip calculating " "expression of mitochondria genes"
                )
        except AttributeError:
            logging.warning(
                "Specified gene column [%s] not found, skip calculating "
                "expression of mitochondria genes",
                gene_name,
            )

    attributes = _get_attributes(adata)
    if list_attr:
        print(_repr_obj(attributes))
        return 0

    conditions, qc_vars, pct_top = _get_filter_conditions(attributes, param, category, subset)

    layer = "counts" if "counts" in adata.layers.keys() else None
    obs_columns = adata.obs.columns
    for qv in qc_vars:
        if f"pct_counts_{qv}" in obs_columns and not force_recalc:
            logging.warning("`pct_counts_%s` exists, not overwriting " "without --force-recalc", qv)
            qc_vars.remove(qv)
    for pt in pct_top:
        if f"pct_counts_in_top_{pt}_genes" in obs_columns and not force_recalc:
            logging.warning("`pct_counts_%s` exists, not overwriting " "without --force-recalc", pt)
            pct_top.remove(pt)
    sc.pp.calculate_qc_metrics(
        adata, layer=layer, qc_vars=qc_vars, percent_top=pct_top, inplace=True
    )
    adata.obs["n_counts"] = adata.obs["total_counts"]
    adata.obs["n_genes"] = adata.obs["n_genes_by_counts"]
    adata.var["n_counts"] = adata.var["total_counts"]
    adata.var["n_cells"] = adata.var["n_cells_by_counts"]

    k_cell = np.ones(len(adata.obs)).astype(bool)
    for cond in conditions["c"]["numerical"]:
        name, vmin, vmax = cond
        attr = adata.obs[name]
        k_cell = k_cell & (attr >= vmin) & (attr <= vmax)

    for cond in conditions["c"]["categorical"]:
        name, values = cond
        attr = getattr(adata.obs, name).astype(str)
        if values[0].startswith("!"):
            values[0] = values[0][1:]
            k_cell = k_cell & (~attr.isin(values))
        else:
            k_cell = k_cell & attr.isin(values)

    k_gene = np.ones(len(adata.var)).astype(bool)
    for cond in conditions["g"]["numerical"]:
        name, vmin, vmax = cond
        attr = adata.var[name]
        k_gene = k_gene & (attr >= vmin) & (attr <= vmax)

    for cond in conditions["g"]["categorical"]:
        name, values = cond
        attr = getattr(adata.var, name).astype(str)
        if values[0].startswith("!"):
            values[0] = values[0][1:]
            k_gene = k_gene & ~(attr.isin(values))
        else:
            k_gene = k_gene & attr.isin(values)

    adata._inplace_subset_obs(k_cell)
    adata._inplace_subset_var(k_gene)


def _get_attributes(adata):
    attributes = {
        "c": {
            "numerical": [],
            "categorical": ["index"],
            "bool": [],
        },
        "g": {
            "numerical": [],
            "categorical": ["index"],
            "bool": [],
        },
    }

    for attr, dtype in adata.obs.dtypes.to_dict().items():
        typ = dtype.kind
        if typ == "O":
            if dtype.name == "category" and dtype.categories.is_boolean():
                attributes["c"]["bool"].append(attr)
            attributes["c"]["categorical"].append(attr)
        elif typ in ("i", "f", "u"):
            attributes["c"]["numerical"].append(attr)
        elif typ == "b":
            attributes["c"]["bool"].append(attr)
            attributes["c"]["categorical"].append(attr)

    for attr, dtype in adata.var.dtypes.to_dict().items():
        typ = dtype.kind
        if typ == "O":
            if dtype.name == "category" and dtype.categories.is_boolean():
                attributes["g"]["bool"].append(attr)
            attributes["g"]["categorical"].append(attr)
        elif typ in ("i", "f", "u"):
            attributes["g"]["numerical"].append(attr)
        elif typ == "b":
            attributes["g"]["bool"].append(attr)
            attributes["g"]["categorical"].append(attr)

    attributes["c"]["numerical"].extend(
        [
            "n_genes",
            "n_counts",
        ]
    )

    for attr in attributes["g"]["bool"]:
        attr2 = "pct_counts_" + attr
        if attr2 not in adata.obs.columns:
            attr2 += "*"
        attributes["c"]["numerical"].append(attr2)

    attributes["g"]["numerical"].extend(
        [
            "n_cells",
            "n_counts",
            "mean_counts",
            "pct_dropout_by_counts",
        ]
    )
    logging.debug(attributes)
    return attributes


def _attributes_exists(name, attributes, dtype):
    cond_cat = ""
    if name.startswith("c:") or name.startswith("g:"):
        cond_cat, _, cond_name = name.partition(":")
        found = int(cond_name in attributes[cond_cat][dtype])
    else:
        cond_name = name
        if cond_name in attributes["c"][dtype]:
            cond_cat += "c"
        if cond_name in attributes["g"][dtype]:
            cond_cat += "g"
        found = len(cond_cat)
    return found, cond_cat, cond_name


def _get_filter_conditions(attributes, param, category, subset):
    conditions = {
        "c": {
            "numerical": [],
            "categorical": [],
            "bool": [],
        },
        "g": {
            "numerical": [],
            "categorical": [],
            "bool": [],
        },
    }
    percent_top_pattern = re.compile(r"^pct_counts_in_top_(?P<n>\d+)_genes$")
    pct_top = []
    qc_vars_pattern = re.compile(r"^pct_counts_(?P<qc_var>\S+)$")
    qc_vars = []

    for name, vmin, vmax in param:
        found, cond_cat, cond_name = _attributes_exists(name, attributes, "numerical")
        pt_match = percent_top_pattern.match(cond_name)
        qv_match = qc_vars_pattern.match(cond_name)
        if found > 1:
            raise ValueError(f'Ambiguous parameter "{name}" found in ' "both cell and gene table")
        if found < 1:
            if pt_match:
                pct_top.append(int(pt_match["n"]))
                cond_cat = "c"
            elif qv_match and qv_match["qc_var"] in attributes["g"]["bool"]:
                qc_vars.append(qv_match["qc_var"])
                cond_cat = "c"
            else:
                raise ValueError(f'Parameter "{name}" unavailable')
        if pt_match or qv_match:
            vmin *= 100
            vmax *= 100
        conditions[cond_cat]["numerical"].append([cond_name, vmin, vmax])

    for name, values in category + subset:
        found, cond_cat, cond_name = _attributes_exists(name, attributes, "categorical")
        if found > 1:
            raise ValueError(f'Ambiguous attribute "{name}" found in ' "both cell and gene table")
        if found < 1:
            raise ValueError(f'Attribute "{name}" unavailable')
        if not isinstance(values, (list, tuple)):
            fh = values
            values = fh.read().rstrip().split("\n")
            fh.close()
        conditions[cond_cat]["categorical"].append((cond_name, values))

    logging.debug((conditions, qc_vars, pct_top))
    return conditions, qc_vars, sorted(pct_top)


def _repr_obj(obj, padding="  ", level=0):
    if isinstance(obj, dict):
        obj_str = "\n".join(
            [
                "\n".join([padding * level + k + ":", _repr_obj(v, level=level + 1)])
                for k, v in obj.items()
            ]
        )
    elif isinstance(obj, (tuple, list, set)):
        obj_str = "\n".join([_repr_obj(elm, level=level) for elm in obj])
    else:
        obj_str = padding * level + repr(obj)
    return obj_str



"""
scanpy paga
"""

import numpy as np
import scanpy as sc
from ._obj_utils import (
    _backup_default_key,
    _delete_backup_key,
    _rename_default_key,
    _set_default_key,
    _restore_default_key,
)

if sc.__version__.startswith("1.4"):
    from scanpy.plotting._tools.scatterplots import plot_scatter
else:
    plot_scatter = sc.pl.embedding


def paga(
    adata,
    use_graph="neighbors",
    key_added=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.paga, for supporting named slot
    """
    _set_default_key(adata.uns, "neighbors", use_graph)
    _backup_default_key(adata.uns, "paga")

    sc.tl.paga(adata, **kwargs)

    _restore_default_key(adata.uns, "neighbors", use_graph)

    if key_added:
        paga_key = f"paga_{key_added}"
        _rename_default_key(adata.uns, "paga", paga_key)
    else:
        _delete_backup_key(adata.uns, "paga")

    return adata


def plot_paga(
    adata,
    use_key="paga",
    basis=None,
    layout=None,
    init_pos=None,
    legend_loc="on data",
    color=None,
    size=None,
    title=None,
    show=None,
    save=None,
    **kwargs,
):
    """Make PAGA plot"""
    if basis is not None and f"X_{basis}" in adata.obsm.keys():
        ax = plot_scatter(
            adata,
            basis=basis,
            color=color,
            legend_loc=legend_loc,
            size=size,
            title=None,
            save=False,
            show=False,
        )

        grouping = adata.uns[use_key]["groups"]
        categories = list(adata.obs[grouping].cat.categories)
        obsm = adata.obsm[f"X_{basis}"]
        group_pos = np.zeros((len(categories), 2))
        for i, label in enumerate(categories):
            offset = 1 if basis.startswith("diffmap") else 0
            _scatter = obsm[adata.obs[grouping] == label, (0 + offset) : (2 + offset)]
            x_pos, y_pos = np.median(_scatter, axis=0)
            group_pos[i] = [x_pos, y_pos]

        _set_default_key(adata.uns, "paga", use_key)
        kwargs["node_size_scale"] = 0
        kwargs["fontsize"] = 1
        kwargs["pos"] = group_pos
        kwargs["color"] = None
        try:
            sc.pl.paga(
                adata,
                ax=ax,
                title=title,
                show=show,
                save=save,
                **kwargs,
            )
        finally:
            _restore_default_key(adata.uns, "paga", use_key)
    else:
        _set_default_key(adata.uns, "paga", use_key)
        try:
            sc.pl.paga(
                adata, layout=layout, init_pos=init_pos, title=title, show=show, save=save, **kwargs
            )
        finally:
            _restore_default_key(adata.uns, "paga", use_key)



import numpy as np
import pandas as pd
import scipy.sparse as sp
from sklearn.preprocessing import normalize
from collections import Counter
from collections import defaultdict
import matplotlib.pyplot as plt
import anndata
import scanpy as sc
from ._diffexp import extract_de_table


class volcano_plot:
    """
    creat volcano plot from anndata
    code from Jonguen Park
    """

    def __init__(
        self, adata, anno_key, comp1, comp2, P=0.1, min_fc=0.5, quick=True, method="ttest"
    ):
        """
        param P :pseudocount for fc calculation
        """
        if method == "ttest":
            from scipy.stats import ttest_ind as test_two_samples
        elif method == "wilcoxon":
            from scipy.stats import wilcoxon as test_two_samples
        self.genelist = adata.raw.var_names
        index1 = adata.obs[anno_key] == comp1
        index2 = adata.obs[anno_key] == comp2

        exp1 = adata.raw[index1].X.todense()
        exp2 = adata.raw[index2].X.todense()

        self.pval = []
        self.fc = []
        for i in range(adata.raw.shape[1]):
            self.fc.append(np.log2((np.mean(exp1[:, i].A1) + P) / (np.mean(exp2[:, i].A1) + P)))
            if quick:
                if np.abs(self.fc[-1]) < min_fc:
                    self.pval.append(1)
                else:
                    self.pval.append(test_two_samples(exp1[:, i].A1, exp2[:, i].A1)[1])
            else:
                self.pval.append(test_two_samples(exp1[:, i].A1, exp2[:, i].A1)[1])

        self.pval = np.array(self.pval)
        from statsmodels.stats.multitest import fdrcorrection

        k_nan = np.isnan(self.pval)
        p = self.pval[~k_nan]
        padj = fdrcorrection(p)[1]
        self.padj = self.pval.copy()
        self.padj[~k_nan] = padj
        self.fc = np.array(self.fc)

    def draw(
        self,
        pvalue_cut=100,
        fc_cut=1,
        adjust_lim=5,
        show=True,
        fontsize=8,
        figsize=(4, 4),
        sided="both",
    ):
        """
        draw volcano plot
        param pvalue_cut :-log10Pvalue for cutoff
        """
        from adjustText import adjust_text

        plt.figure(figsize=figsize)

        xpos = np.array(self.fc)
        ypos = -np.log10(np.array(self.padj))
        ypos[ypos == np.inf] = np.max(ypos[ypos != np.inf])

        if sided == "upper":
            sig = (xpos > fc_cut) & (ypos > pvalue_cut)
        elif sided == "lower":
            sig = (xpos < fc_cut) & (ypos > pvalue_cut)
        else:
            sig = (np.abs(xpos) > fc_cut) & (ypos > pvalue_cut)

        plt.scatter(xpos, ypos, s=1, color="k", rasterized=True)
        plt.scatter(xpos[sig], ypos[sig], s=3, color="red", rasterized=True)

        texts = []
        for i, gene in enumerate(self.genelist[sig]):
            texts.append(plt.text(xpos[sig][i], ypos[sig][i], gene, fontsize=fontsize))

        adjust_text(texts, only_move={"texts": "xy"}, lim=adjust_lim)
        if show:
            plt.show()


def calc_marker_stats(ad, groupby, genes=None, use_rep="raw", inplace=False, partial=False):
    if ad.obs[groupby].dtype.name != "category":
        raise ValueError('"%s" is not categorical' % groupby)
    n_grp = ad.obs[groupby].cat.categories.size
    if n_grp < 2:
        raise ValueError('"%s" must contain at least 2 categories' % groupby)
    if use_rep == "raw":
        X = ad.raw.X
        var_names = ad.raw.var_names.values
    else:
        X = ad.X
        var_names = ad.var_names.values
    if not sp.issparse(X):
        X = sp.csr_matrix(X)
    if genes:
        v_idx = var_names.isin(genes)
        X = X[:, v_idx]
        var_names = var_names[v_idx]

    X = normalize(X, norm="max", axis=0)
    k_nonzero = X.sum(axis=0).A1 > 0
    X = X[:, np.where(k_nonzero)[0]]
    var_names = var_names[k_nonzero]

    n_var = var_names.size
    x = np.arange(n_var)

    grp_indices = {k: g.index.values for k, g in ad.obs.reset_index().groupby(groupby, sort=False)}

    frac_df = pd.DataFrame(
        {k: (X[idx, :] > 0).mean(axis=0).A1 for k, idx in grp_indices.items()}, index=var_names
    )
    mean_df = pd.DataFrame(
        {k: X[idx, :].mean(axis=0).A1 for k, idx in grp_indices.items()}, index=var_names
    )

    if partial:
        stats_df = None
    else:
        frac_order = np.apply_along_axis(np.argsort, axis=1, arr=frac_df.values)
        y1 = frac_order[:, n_grp - 1]
        y2 = frac_order[:, n_grp - 2]
        y3 = frac_order[:, n_grp - 3] if n_grp > 2 else y2
        top_frac_grps = frac_df.columns.values[y1]
        top_fracs = frac_df.values[x, y1]
        frac_diffs = top_fracs - frac_df.values[x, y2]
        max_frac_diffs = top_fracs - frac_df.values[x, y3]

        mean_order = np.apply_along_axis(np.argsort, axis=1, arr=mean_df.values)
        y1 = mean_order[:, n_grp - 1]
        y2 = mean_order[:, n_grp - 2]
        y3 = mean_order[:, n_grp - 3] if n_grp > 2 else y2
        top_mean_grps = mean_df.columns.values[y1]
        top_means = mean_df.values[x, y1]
        mean_diffs = top_means - mean_df.values[x, y2]
        max_mean_diffs = top_means - mean_df.values[x, y3]

        stats_df = pd.DataFrame(
            {
                "top_frac_group": top_frac_grps,
                "top_frac": top_fracs,
                "frac_diff": frac_diffs,
                "max_frac_diff": max_frac_diffs,
                "top_mean_group": top_mean_grps,
                "top_mean": top_means,
                "mean_diff": mean_diffs,
                "max_mean_diff": max_mean_diffs,
            },
            index=var_names,
        )
        stats_df["top_frac_group"] = stats_df["top_frac_group"].astype(
            pd.CategoricalDtype(categories=list(ad.obs[groupby].cat.categories), ordered=True)
        )

    if inplace:
        if use_rep == "raw":
            ad.raw.varm[f"frac_{groupby}"] = frac_df
            ad.raw.varm[f"mean_{groupby}"] = mean_df
            if not partial:
                ad.raw.var = pd.concat([ad.raw.var, stats_df], axis=1)
        else:
            ad.varm[f"frac_{groupby}"] = frac_df
            ad.varm[f"mean_{groupby}"] = mean_df
            if not partial:
                ad.var = pd.concat([ad.raw.var, stats_df], axis=1)
    else:
        return frac_df, mean_df, stats_df


def filter_marker_stats(
    ad,
    use_rep="raw",
    min_frac_diff=0.1,
    min_mean_diff=0.1,
    max_next_frac=0.9,
    max_next_mean=0.95,
    single=False,
    how="or",
):
    columns = [
        "top_frac_group",
        "top_frac",
        "frac_diff",
        "max_frac_diff",
        "top_mean_group",
        "top_mean",
        "mean_diff",
        "max_mean_diff",
    ]
    if isinstance(ad, anndata.AnnData):
        stats_df = ad.raw.var[columns] if use_rep == "raw" else ad.var[columns]
    elif isinstance(ad, pd.DataFrame):
        stats_df = ad[columns]
    else:
        raise ValueError("Invalid input, must be an AnnData or DataFrame")
    frac_diff = stats_df.frac_diff if single else stats_df.max_frac_diff
    mean_diff = stats_df.mean_diff if single else stats_df.max_mean_diff
    same_group = stats_df.top_frac_group == stats_df.top_mean_group
    meet_frac_requirement = (frac_diff >= min_frac_diff) & (
        stats_df.top_frac - frac_diff <= max_next_frac
    )
    meet_mean_requirement = (mean_diff >= min_mean_diff) & (
        stats_df.top_mean - mean_diff <= max_next_mean
    )
    if how == "or":
        filtered = stats_df.loc[same_group & (meet_frac_requirement | meet_mean_requirement)]
    else:
        filtered = stats_df.loc[same_group & (meet_frac_requirement & meet_mean_requirement)]
    if single:
        filtered = filtered.sort_values(
            ["top_frac_group", "mean_diff", "frac_diff"], ascending=[True, False, False]
        )
    else:
        filtered = filtered.sort_values(
            ["top_frac_group", "mean_diff", "frac_diff"], ascending=[True, False, False]
        )
    filtered["top_frac_group"] = filtered["top_frac_group"].astype("category")
    filtered["top_frac_group"].cat.reorder_categories(
        list(stats_df["top_frac_group"].cat.categories), inplace=True
    )
    return filtered


def top_markers(df, top_n=5, groupby="top_frac_group"):
    return df.groupby(groupby).head(top_n).index.to_list()


def test_markers(ad, mks, groupby, n_genes=100, use_raw=True, **kwargs):
    genes = top_markers(mks, top_n=n_genes)
    aux_ad = anndata.AnnData(
        X=ad.raw.X if use_raw else ad.X,
        obs=ad.obs.copy(),
        var=ad.raw.var.copy() if use_raw else ad.var.copy(),
    )
    aux_ad = aux_ad[:, genes].copy()
    sc.tl.rank_genes_groups(aux_ad, groupby=groupby, n_genes=n_genes, use_raw=False, **kwargs)
    de_tbl = extract_de_table(aux_ad.uns["rank_genes_groups"])
    return (
        mks.reset_index()
        .rename(columns={"index": "genes", "top_frac_group": "cluster"})
        .merge(de_tbl[["cluster", "genes", "logfoldchanges", "pvals", "pvals_adj"]], how="left")
    )



"""
Remove doublets
"""

import numpy as np
import scipy.sparse as sp
from scipy.stats import norm
import pandas as pd
import scanpy as sc
import anndata
from ._utils import lognorm_to_counts, _find_rep


def _calculate_z(x, upper_mad_only=True):
    med = np.median(x)
    if upper_mad_only:
        mad = np.median(x[x > med] - med) * 1.4826
    else:
        mad = np.median(np.abs(x - med)) * 1.4826
    return (x - med) / mad


def _test_outlier(x, upper_mad_only=True):
    from statsmodels.stats.multitest import multipletests

    z = _calculate_z(x, upper_mad_only=upper_mad_only)
    pvals = 1 - norm.cdf(z, loc=0, scale=1)
    bh_pvals = multipletests(pvals, method="fdr_bh")[1]
    return pvals, bh_pvals, z


def run_scrublet(
    adata,
    resolution_function=None,
    use_rep="X",
    reverse_lognorm=False,
    inplace=True,
    ignore_error=True,
):
    if resolution_function is None:
        resolution_function = lambda x: np.maximum(np.maximum(np.log10(x) - 1, 0) ** 2, 0.1)

    old_verbosity = sc.settings.verbosity
    sc.settings.verbosity = 1

    use_calculated = False
    X = _find_rep(adata, use_rep=use_rep, return_var_names=False)

    if not sp.issparse(X):
        X = sp.csr_matrix(X)
    if reverse_lognorm:
        X = lognorm_to_counts(X)

    import scrublet as scr

    try:
        scrub = scr.Scrublet(X)
        ds, _ = scrub.scrub_doublets(verbose=False)
        ds_z = _calculate_z(ds)

        adata_copy = anndata.AnnData(X=X, obs=adata.obs)
        sc.pp.filter_genes(adata_copy, min_cells=3)
        sc.pp.normalize_total(adata_copy, target_sum=1e4)
        sc.pp.log1p(adata_copy)
        sc.pp.highly_variable_genes(
            adata_copy, min_mean=0.0125, max_mean=3, min_disp=0.5, subset=True
        )
        sc.pp.scale(adata_copy, zero_center=False)
        sc.pp.pca(adata_copy, svd_solver="arpack", zero_center=False)
        sc.pp.neighbors(adata_copy, n_pcs=30)
        sc.tl.leiden(adata_copy, resolution=1)
        for clst in np.unique(adata_copy.obs["leiden"]):
            clst_size = sum(adata_copy.obs["leiden"] == clst)
            sc.tl.leiden(
                adata_copy,
                restrict_to=("leiden", [clst]),
                resolution=resolution_function(clst_size),
                key_added="leiden_R",
            )
            adata_copy.obs["leiden"] = adata_copy.obs["leiden_R"]
        clst_meds = []
        for clst in np.unique(adata_copy.obs["leiden"]):
            k = adata_copy.obs["leiden"] == clst
            clst_med = np.median(ds[k])
            adata_copy.obs.loc[k, "cluster_scrublet_score"] = clst_med
            clst_meds.append(clst_med)
        clst_meds = np.array(clst_meds)
        pvals, bh_pvals, z_scores = _test_outlier(clst_meds)
        for i, clst in enumerate(np.unique(adata_copy.obs["leiden"])):
            k = adata_copy.obs["leiden"] == clst
            adata_copy.obs.loc[k, "pval"] = pvals[i]
            adata_copy.obs.loc[k, "bh_pval"] = bh_pvals[i]
        sc.settings.verbosity = old_verbosity

        df = adata_copy.obs[["cluster_scrublet_score", "bh_pval"]].copy()
        del adata_copy

        df["scrublet_score"] = ds
        df["scrublet_score_z"] = ds_z
        df["scrublet_done"] = True

    except ValueError:
        if ignore_error:
            n = adata.n_obs
            ones = np.ones(n)
            zeros = np.zeros(n)
            df = pd.DataFrame(
                {
                    "cluster_scrublet_score": zeros,
                    "bh_pval": ones,
                    "scrublet_score": zeros,
                    "scrublet_score_z": zeros,
                },
                index=adata.obs_names,
            )
        df["scrublet_done"] = False

    if inplace:
        adata.obs["scrublet_score"] = df["scrublet_score"].values
        adata.obs["scrublet_score_z"] = df["scrublet_score_z"].values
        adata.obs["cluster_scrublet_score"] = df["cluster_scrublet_score"].values
        adata.obs["bh_pval"] = df["bh_pval"].values
        adata.obs["scrublet_done"] = df["scrublet_done"].values
    else:
        return df



"""
scanpy diffmap
"""

import scanpy as sc
from ._obj_utils import (
    _set_default_key,
    _restore_default_key,
    _backup_obsm_key,
    _delete_obsm_backup_key,
    _rename_obsm_key,
    write_embedding,
)


def diffmap(
    adata,
    use_graph="neighbors",
    key_added=None,
    export_embedding=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.diffmap, for supporting named slot
    """
    _set_default_key(adata.uns, "neighbors", use_graph)
    _backup_obsm_key(adata, "X_diffmap")

    sc.tl.diffmap(adata, **kwargs)

    _restore_default_key(adata.uns, "neighbors", use_graph)

    diffmap_key = "X_diffmap"
    if key_added:
        diffmap_key = f"{diffmap_key}_{key_added}"
        _rename_obsm_key(adata, "X_diffmap", diffmap_key)
    else:
        _delete_obsm_backup_key(adata, "X_diffmap")

    if export_embedding is not None:
        write_embedding(adata, diffmap_key, export_embedding, key_added=key_added)



"""
scanpy leiden
"""

import re
import numpy as np
import scanpy as sc
from ._obj_utils import write_cluster


def leiden(
    adata,
    resolution,
    flavor=None,
    use_graph=None,
    key_added=None,
    export_cluster=None,
    legacy=False,
    **kwargs,
):
    """
    Wrapper function for sc.tl.leiden, for supporting multiple resolutions.
    """
    keys = []
    if "restrict_to" in kwargs and not (
        isinstance(kwargs["restrict_to"], (list, tuple))
        and len(kwargs["restrict_to"]) == 2
        and kwargs["restrict_to"][0]
    ):
        del kwargs["restrict_to"]
    adj_mat = None
    if use_graph:
        if use_graph not in adata.uns:
            raise KeyError(f'"{use_graph}" is not a valid key of `.uns`.')
        if sc.__version__.startswith("1.4") or legacy:
            adj_mat = adata.uns[use_graph]["connectivities"]
        else:
            ckey = adata.uns[use_graph]["connectivities_key"]
            adj_mat = adata.obsp[ckey]
    if not isinstance(resolution, (list, tuple)):
        if key_added is not None and not key_added.startswith("leiden_"):
            key_added = f"leiden_{key_added}"
        elif key_added is None:
            key_added = "leiden"
        if flavor == "rapids":
            kwargs["flavor"] = "rapids"
        sc.tl.leiden(adata, resolution=resolution, adjacency=adj_mat, key_added=key_added, **kwargs)
        keys.append(key_added)
    else:
        for i, res in enumerate(resolution):
            res_key = str(res).replace(".", "_")
            if key_added is None:
                graph_key = ("_" + use_graph) if use_graph else ""
                key = f"leiden{graph_key}_r{res_key}"
            elif not isinstance(key_added, (list, tuple)):
                key = f"leiden_{key_added}_r{res_key}"
            elif len(key_added) == len(resolution):
                key = key_added[i]
            else:
                raise ValueError(
                    "`key_added` can only be None, a scalar, or an "
                    "iterable of the same length as `resolution`."
                )
            keys.extend(
                leiden(
                    adata,
                    resolution=res,
                    flavor=flavor,
                    use_graph=use_graph,
                    key_added=key,
                    **kwargs,
                )
            )

    if export_cluster:
        write_cluster(adata, keys, export_cluster)

    return keys


def leiden_shredding(
    adata,
    resolution=[0.5, 0.1],
    use_graph=None,
    level1_groups=None,
    key_added=None,
    min_cluster_size=10,
    max_cluster_size=1000,
    **kwargs,
):
    # resolution_function = lambda x: np.maximum(np.maximum(np.log10(x)-1, 0)**2, 0.1)
    if not isinstance(resolution, (list, tuple)):
        resolution = [resolution, resolution]
    adj_mat = None
    if use_graph:
        if use_graph not in adata.uns:
            raise KeyError(f'"{use_graph}" is not a valid key of `.uns`.')
        if sc.__version__.startswith("1.4"):
            adj_mat = adata.uns[use_graph]["connectivities"]
        else:
            ckey = adata.uns[use_graph]["connectivities_key"]
            adj_mat = adata.obsp[ckey]
    if key_added is not None and not key_added.startswith("leiden_"):
        key_added = f"leiden_{key_added}"
    elif key_added is None:
        key_added = "leiden"
    if level1_groups and level1_groups in adata.obs.columns:
        adata.obs[key_added] = adata.obs[level1_groups].copy()
    else:
        sc.tl.leiden(
            adata, adjacency=adj_mat, resolution=resolution[0], key_added=key_added, **kwargs
        )
    level1_clusters = np.unique(adata.obs[key_added])
    for clst in level1_clusters:
        clst_size = sum(adata.obs[key_added] == clst)
        if clst_size <= min_cluster_size:
            continue
        if max_cluster_size and clst_size <= max_cluster_size:
            continue
        sc.tl.leiden(
            adata,
            restrict_to=(key_added, [clst]),
            key_added="leiden_temp",
            adjacency=adj_mat,
            resolution=resolution[1] * max(1, np.sqrt(clst_size / min_cluster_size)),
            **kwargs,
        )
        adata.obs[key_added] = adata.obs["leiden_temp"]
    del adata.obs["leiden_temp"]


def leiden_shredding2(
    adata,
    adjacency=None,
    resolution=0.5,
    initial_groupby=None,
    restrict_to=None,
    key_added=None,
    level=0,
    min_cluster_size=20,
    max_cluster_size=200,
    **kwargs,
):
    tmp_key = f"{key_added}_{level}"
    if initial_groupby and initial_groupby in adata.obs.columns:
        adata.obs[tmp_key] = adata.obs[initial_groupby].copy()
    else:
        sc.tl.leiden(
            adata,
            adjacency=adjacency,
            restrict_to=restrict_to,
            resolution=resolution,
            key_added=tmp_key,
            **kwargs,
        )
    current_clusters = np.unique(adata.obs[tmp_key])
    if restrict_to is not None:
        _, parent_group = restrict_to
        parent_group = parent_group[0]
        clst_pattern = re.compile(f"{re.escape(parent_group)}(,[0-9]+)+$")
        current_clusters = [clst for clst in current_clusters if clst_pattern.match(clst)]
    clst_sizes =  adata.obs[tmp_key].value_counts().loc[current_clusters]
    if len(current_clusters) < 2 and clst_sizes.max() > max_cluster_size:
        # require re-do with increased resolution
        return 1
    if (
        min_cluster_size is not None
        and (level > 0 or initial_groupby is None)
        and clst_sizes.min() < min_cluster_size
        and (clst_sizes < min_cluster_size).sum() > 1
        and len(current_clusters) > 2
    ):
        # require re-do with decreased resolution
        return -1
    for clst in current_clusters:
        clst_size = sum(adata.obs[tmp_key] == clst)
        res = resolution
        if max_cluster_size is not None and clst_size > max_cluster_size:
            # require another round of leiden
            max_res, min_res = None, None
            while True:
                ret = leiden_shredding2(
                    adata,
                    adjacency=adjacency,
                    resolution=res,
                    restrict_to=(tmp_key, [clst]),
                    key_added=key_added,
                    min_cluster_size=min_cluster_size,
                    max_cluster_size=max_cluster_size,
                    level=level + 1,
                    **kwargs
                )
                if ret == 0:
                    break

                if ret > 0:
                    if max_res is None:
                        next_res = res + 0.1 if res >= 0.1 else res * 2
                    else:
                        next_res = (res + max_res) / 2
                    min_res = res
                else:
                    if min_res is None:
                        next_res = res - 0.1 if res >= 0.2 else res / 2
                    else:
                        next_res = (res + min_res) / 2
                    max_res = res

                # Don't search forever
                if np.abs(next_res - res) < 0.01:
                    print("stop search")
                    break

                res = next_res

            next_tmp_key = f"{key_added}_{level + 1}"
            adata.obs[tmp_key] = adata.obs[next_tmp_key]
            del adata.obs[next_tmp_key]

    if level == 0:
        adata.obs[key_added] = adata.obs[tmp_key]
        del adata.obs[tmp_key]

    return 0



"""
scanpy neighbors
"""

import scanpy as sc
from ._obj_utils import (
    _backup_default_key,
    _delete_backup_key,
    _rename_default_key,
)


def neighbors(adata, n_neighbors=15, key_added=None, **kwargs):
    """
    Wrapper function for sc.pp.neighbors(), for supporting multiple n_neighbors
    """
    if not isinstance(n_neighbors, (list, tuple)):
        _backup_default_key(adata.uns, "neighbors")

        if sc.__version__.startswith("1.4"):
            sc.pp.neighbors(adata, n_neighbors=n_neighbors, **kwargs)

            if key_added:
                nb_key = f"neighbors_{key_added}"
                _rename_default_key(adata.uns, "neighbors", nb_key)
            else:
                _delete_backup_key(adata.uns, "neighbors")
        else:
            if key_added:
                nb_key = f"neighbors_{key_added}"
            sc.pp.neighbors(adata, n_neighbors=n_neighbors, key_added=nb_key, **kwargs)
    else:
        for i, n_nb in enumerate(n_neighbors):
            if key_added is None:
                graph_key = f"k{n_nb}"
            elif not isinstance(key_added, (list, tuple)):
                graph_key = f"{key_added}_k{n_nb}"
            elif len(key_added) == len(n_neighbors):
                graph_key = key_added[i]
            else:
                raise ValueError(
                    "`key_added` can only be None, a scalar, or an "
                    "iterable of the same length as `n_neighbors`."
                )
            neighbors(
                adata,
                n_neighbors=n_nb,
                key_added=graph_key,
                **kwargs,
            )



"""
scanpy tsne
"""

import scanpy as sc
from ._obj_utils import (
    _backup_obsm_key,
    _rename_obsm_key,
    _delete_obsm_backup_key,
    write_embedding,
)


def tsne(
    adata,
    key_added=None,
    random_state=0,
    export_embedding=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.tsne, for supporting named slot of tsne embeddings
    """
    if not isinstance(random_state, (list, tuple)):
        _backup_obsm_key(adata, "X_tsne")

        sc.tl.tsne(adata, random_state=random_state, **kwargs)

        tsne_key = "X_tsne"
        if key_added:
            tsne_key = f"X_tsne_{key_added}"
            _rename_obsm_key(adata, "X_tsne", tsne_key)
        else:
            _delete_obsm_backup_key(adata, "X_tsne")

        if export_embedding is not None:
            write_embedding(adata, tsne_key, export_embedding, key_added=key_added)
    else:
        for i, rseed in enumerate(random_state):
            if key_added is None:
                tsne_key = f"r{rseed}"
            elif not isinstance(key_added, (list, tuple)):
                tsne_key = f"{key_added}_r{rseed}"
            elif len(key_added) == len(random_state):
                tsne_key = key_added[i]
            else:
                raise ValueError(
                    "`key_added` can only be None, a scalar, or "
                    "an iterable of the same length as "
                    "`random_state`."
                )
            tsne(
                adata,
                key_added=tsne_key,
                random_state=rseed,
                **kwargs,
            )



"""
scanpy fdg
"""

import scanpy as sc
from ._obj_utils import (
    _backup_obsm_key,
    _delete_obsm_backup_key,
    _rename_obsm_key,
    write_embedding,
)


def fdg(
    adata,
    use_graph="neighbors",
    layout="fa",
    key_added=None,
    random_state=0,
    export_embedding=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.draw_graph, for supporting named slot of fdg
    embeddings.
    """
    adj_mat = None
    if use_graph != "neighbors":
        if use_graph not in adata.uns.keys():
            raise KeyError(f'"{use_graph}" is not a valid key of `.uns`.')
        if "connectivities" in adata.uns[use_graph].keys():
            adj_mat = adata.uns[use_graph]["connectivities"]
        elif getattr(adata, "obsp") and f"{use_graph}_connectivities" in adata.obsp.keys():
            adj_mat = adata.obsp[f"{use_graph}_connectivities"]
        else:
            raise KeyError(f'connectivities graph not found for "{use_graph}".')

    _backup_obsm_key(adata, f"X_draw_graph_{layout}")

    sc.tl.draw_graph(
        adata,
        layout=layout,
        random_state=random_state,
        adjacency=adj_mat,
        **kwargs,
    )

    fdg_key = f"X_draw_graph_{layout}"
    if key_added:
        fdg_key = f"X_draw_graph_{layout}_{key_added}"
        _rename_obsm_key(adata, f"X_draw_graph_{layout}", fdg_key)
    else:
        _delete_obsm_backup_key(adata, f"X_draw_graph_{layout}")

    if export_embedding is not None:
        write_embedding(adata, fdg_key, export_embedding, key_added=key_added)



"""
scanpy hvg
"""

import numpy as np
import pandas as pd
import scanpy as sc
from ._utils import sc_warn


def hvg(
    adata,
    mean_limits=(0.0125, 3),
    disp_limits=(0.5, float("inf")),
    subset=False,
    by_batch=None,
    n_hvg=None,
    **kwargs,
):
    """
    Wrapper function for sc.highly_variable_genes(), mainly to support searching
    by batch and pooling.
    """

    # Check for n_top_genes beeing greater than the total genes

    if "n_top_genes" in kwargs and kwargs["n_top_genes"] is not None:
        kwargs["n_top_genes"] = min(adata.n_vars, kwargs["n_top_genes"])

    if by_batch and isinstance(by_batch, (list, tuple)) and by_batch[0]:
        kwargs["batch_key"] = by_batch[0]

    sc.pp.highly_variable_genes(
        adata,
        min_disp=disp_limits[0],
        max_disp=disp_limits[1],
        min_mean=mean_limits[0],
        max_mean=mean_limits[1],
        inplace=True,
        **kwargs,
    )

    if by_batch and isinstance(by_batch, (list, tuple)) and by_batch[0]:
        batch_name = by_batch[0]
        min_n = by_batch[1]
        k_hvg = adata.var.highly_variable_nbatches
        if n_hvg is not None and isinstance(n_hvg, int):
            n_cum = k_hvg.value_counts().sort_index(ascending=False).cumsum()
            if np.all(n_cum < n_hvg):
                min_n = 1
            else:
                min_n = n_cum[n_cum > n_hvg].index[0]
        sc_warn(f"n_hvg = {(k_hvg >= min_n).sum()} found in at least {min_n} batches")
        if subset:
            adata = adata[:, k_hvg.values >= min_n]
        else:
            adata.var["highly_variable"] = k_hvg >= min_n



"""
scanpy dpt
"""

import numpy as np
import scanpy as sc
from ._obj_utils import (
    _set_default_key,
    _restore_default_key,
    _backup_default_key,
    _delete_backup_key,
    _rename_default_key,
    _set_obsm_key,
    _restore_obsm_key,
)


def dpt(
    adata,
    root=None,
    use_graph="neighbors",
    use_diffmap="X_diffmap",
    key_added=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.dpt, for support named slot
    """
    if root is None or not (isinstance(root, (list, tuple)) and len(root) == 2):
        root = (None, None)
    if "iroot" not in adata.uns.keys() and root[0] is None:
        raise ValueError(
            "Annotate your data with root cell first, i.e. "
            'boolean vector `.uns["iroot"]` is required.'
        )
    if root[0] is not None:
        adata.uns["iroot"] = np.random.choice(np.flatnonzero(adata.obs[root[0]] == root[1]))

    _set_default_key(adata.uns, "neighbors", use_graph)
    _set_obsm_key(adata, "X_diffmap", use_diffmap)

    _backup_default_key(adata.obs, "dpt_pseudotime")

    sc.tl.dpt(adata, **kwargs)

    _restore_default_key(adata.uns, "neighbors", use_graph)
    _restore_obsm_key(adata, "X_diffmap", use_diffmap)

    if key_added:
        dpt_key = f"dpt_pseudotime_{key_added}"
        _rename_default_key(adata.obs, "dpt_pseudotime", dpt_key)
    else:
        _delete_backup_key(adata.obs, "dpt_pseudotime")



"""
Transfer annotation functions
"""

import os
import sys
import numpy as np
import numpy_groupies as npg
import scipy.sparse as sp
import pandas as pd
import joblib
from copy import deepcopy
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import adjusted_rand_score
from sklearn.exceptions import ConvergenceWarning
from sklearn.base import ClassifierMixin

try:
    from sklearn.utils.testing import ignore_warnings
except Exception:
    from sklearn.utils._testing import ignore_warnings


@ignore_warnings(category=ConvergenceWarning)
def LR_train(
    adata,
    groupby,
    use_rep="raw",
    use_hvg=False,
    use_pseudobulk=False,
    max_pass=20,
    downsample_to=None,
    save=None,
    model=None,
    **kwargs,
):
    if downsample_to is not None:
        if (
            isinstance(downsample_to, (tuple, list))
            and isinstance(downsample_to[0], int)
            and isinstance(downsample_to[1], str)
        ):
            from ._utils import subsample

            ad = subsample(adata, 1, groupby=groupby, max_n=downsample_to[0])
            ad.write(downsample_to[1], compression="lzf")
        else:
            raise ValueError("`downsample_to` needs to be (int[N], str[h5ad])")
    else:
        ad = adata

    groupby_var = ad.obs[groupby].cat.remove_unused_categories()
    Y = groupby_var.astype(str)

    if use_rep == "raw":
        X = ad.raw.X
        features = ad.raw.var_names.values
    elif use_rep == "X":
        X = ad.X
        features = ad.var_names.values
        if use_hvg and "highly_variable" in ad.var.keys():
            k_hvg = ad.var["highly_variable"].values
            X = X[:, k_hvg]
            features = features[k_hvg]
    elif use_rep in ad.obsm.keys():
        X = ad.obsm[use_rep]
        features = np.array([f"V{i+1}" for i in range(X.shape[1])])
    else:
        raise KeyError(f"{use_rep}: invalid <use_rep>")

    if use_pseudobulk:
        if sp.issparse(X):
            summarised = np.zeros((Y.unique().size, X.shape[1]))
            for i, grp in enumerate(groupby_var.cat.categories):
                k_grp = np.where(Y == grp)[0]
                summarised[i] = np.mean(X[k_grp, :], axis=0)
            X = summarised
        else:
            X = npg.aggregate(groupby_var.cat.codes, X, axis=0, func="mean")
        Y = groupby_var.cat.categories.values

    lr = (
        model
        if model is not None
        else LogisticRegression(
            penalty="l2", C=0.1, solver="saga", warm_start=True, n_jobs=-1, **kwargs
        )
    )
    n_pass = 0
    while n_pass < max_pass:
        lr.fit(X, Y)
        n_pass += 1
        if lr.n_iter_ < 100:
            break
    if lr.n_iter_ >= 100:
        print("training of LR model failed to converge", file=sys.stderr)
    lr.features = features
    if save:
        outdir = os.path.dirname(save)
        if outdir:
            os.makedirs(outdir, exist_ok=True)
        joblib.dump(lr, save)
    return lr


def LR_predict(
    adata,
    model,
    use_rep="raw",
    use_pseudobulk=False,
    groupby=None,
    feature=None,
    key_added=None,
    truth=None,
    return_predict=False,
    min_prob=None,
):
    if use_rep == "raw":
        X = adata.raw.X
        features = adata.raw.var_names if feature is None else adata.raw.var[feature]
    elif use_rep == "X":
        X = adata.X
        features = adata.var_names if feature is None else adata.var[feature]
    elif use_rep in adata.obsm.keys():
        X = adata.obsm[use_rep]
        features = np.array([f"V{i+1}" for i in range(X.shape[1])])
    else:
        raise KeyError(f"{use_rep}: invalid <use_rep>")
    features = pd.Series(features)

    if isinstance(model, str) and os.path.exists(model):
        lr = joblib.load(model)
    elif isinstance(model, ClassifierMixin):
        lr = deepcopy(model)
    else:
        raise ValueError(f"{model}: invalid LR model")
    if getattr(lr, "features", None) is None:
        if lr.n_features_in_ == features.size:
            lr.features = features.values
        else:
            raise ValueError(f"{model}: LR model has no feature names and unmatched size")

    k_x = features.isin(list(lr.features))
    print(f"{k_x.sum()} features used for prediction", file=sys.stderr)
    k_x_idx = np.where(k_x)[0]
    X = X[:, k_x_idx]
    features = features[k_x]

    ad_ft = (
        pd.DataFrame(features.values, columns=["ad_features"])
        .reset_index()
        .rename(columns={"index": "ad_idx"})
    )
    lr_ft = (
        pd.DataFrame(lr.features, columns=["lr_features"])
        .reset_index()
        .rename(columns={"index": "lr_idx"})
    )
    lr_idx = (
        lr_ft.merge(ad_ft, left_on="lr_features", right_on="ad_features")
        .sort_values(by="ad_idx")
        .lr_idx.values
    )

    lr.n_features_in_ = lr_idx.size
    lr.features = lr.features[lr_idx]
    lr.coef_ = lr.coef_[:, lr_idx]

    if use_pseudobulk:
        if not groupby or groupby not in adata.obs.columns:
            raise ValueError("missing or invalid `groupby`")
        groupby_var = adata.obs[groupby].cat.remove_unused_categories()
        summarised = np.zeros((groupby_var.cat.categories.size, X.shape[1]))
        for i, grp in enumerate(groupby_var.cat.categories):
            k_grp = np.where(groupby_var == grp)[0]
            if sp.issparse(X):
                summarised[i] = np.mean(X[k_grp, :], axis=0)
            else:
                summarised[i] = np.mean(X[k_grp, :], axis=0, keepdims=True)
        X = summarised

    ret = {}
    Y_predict = lr.predict(X)
    Y_prob = lr.predict_proba(X)
    max_Y_prob = Y_prob.max(axis=1)
    if use_pseudobulk:
        tmp_groupby = adata.obs[groupby].astype(str)
        tmp_prob = np.zeros(tmp_groupby.size)
        tmp_predict = tmp_prob.astype(str)
        for i, ct in enumerate(adata.obs[groupby].cat.categories):
            tmp_prob[tmp_groupby == ct] = max_Y_prob[i]
            tmp_predict[tmp_groupby == ct] = Y_predict[i]
        max_Y_prob = tmp_prob
        Y_predict = tmp_predict
    if min_prob is not None:
        Y_predict[max_Y_prob < min_prob] = "Uncertain"
    if key_added:
        adata.obs[key_added] = Y_predict
        adata.obs[key_added] = adata.obs[key_added].astype("category")
        adata.obs[f"{key_added}_prob"] = max_Y_prob
    ret["label"] = Y_predict
    ret["prob"] = pd.DataFrame(
        Y_prob,
        index=adata.obs[groupby].cat.categories if use_pseudobulk else adata.obs_names,
        columns=lr.classes_,
    )

    if truth:
        Y_truth = adata.obs[truth].astype(str)
        ret["accuracy"] = (Y_predict == Y_truth).sum() / Y_predict.size
        ret["adjusted_rand_score"] = adjusted_rand_score(Y_truth, Y_predict)

    if return_predict:
        return ret


def annotate(adata, groupby, label, normalise_label=True, threshold=0.85, max_entry=None):
    """Annotate a clustering based on a matrix of values

    + groupby: the clustering to be annotated
    + label : cell-level annotation
    """
    annot_mat = pd.crosstab(adata.obs[groupby], adata.obs[label])
    group_size = annot_mat.sum(axis=1).values
    group_prop = annot_mat / group_size[:, np.newaxis]
    if normalise_label:
        label_size = group_prop.sum(axis=0).values
        label_prop = group_prop / label_size[np.newaxis, :]
    else:
        label_prop = group_prop
    v_max = label_prop.values.max(axis=1)

    annot_dict = {}
    for i, row in label_prop.iterrows():
        idx = np.where(row.values > row.values.max() * threshold)[0]
        od = np.argsort(row.values[idx])
        if max_entry is not None:
            max_entry = min(max_entry, len(idx))
            entries = row[idx[od]][0:max_entry]
        else:
            entries = row[idx[od]]
        gl = ";".join(entries.index.values)
        if gl not in annot_dict:
            annot_dict[gl] = []
        annot_dict[gl].append(i)
    return annot_dict


def get_top_LR_features(model, group, n_top=10):
    return (
        pd.DataFrame(
            {"gene": model.features, "coef": model.coef_[model.classes_ == group, :].flatten()}
        )
        .sort_values("coef", ascending=False)
        .head(n_top)
    )



"""
Plotting related functions
"""

from collections import defaultdict
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sn
from matplotlib import rcParams
from matplotlib.colors import LinearSegmentedColormap, ListedColormap, Normalize, to_hex
from matplotlib.lines import Line2D
import anndata
import scanpy as sc
from ._diffexp import extract_de_table
from ._utils import pseudo_bulk, summarise_expression_by_group, sc_warn, cross_table

if sc.__version__.startswith("1.4"):
    from scanpy.plotting._tools.scatterplots import plot_scatter
else:
    plot_scatter = sc.pl.embedding

rcParams["pdf.fonttype"] = 42
rcParams["ps.fonttype"] = 42


def expression_colormap(background_level=0.01):
    """Returns a nice color map for highlighting gene expression"""
    background_nbin = int(100 * background_level)
    reds = plt.cm.Reds(np.linspace(0, 1, 100 - background_nbin))
    greys = plt.cm.Greys_r(np.linspace(0.7, 0.8, background_nbin))
    palette = np.vstack([greys, reds])
    return LinearSegmentedColormap.from_list("expression", palette)


def make_palette(n, cmap=None, hide_first=False, hide_last=False, hide_color="#E9E9E910"):
    """Returns a color palette with specified number of colors"""
    i = int(hide_first)
    j = int(hide_last)
    if cmap is None:
        if sc.__version__.startswith("1.4"):
            sc_default_10 = rcParams["axes.prop_cycle"].by_key()["color"]
            sc_default_26 = sc.plotting.palettes.default_26
            sc_default_64 = sc.plotting.palettes.default_64
            palette = (
                sc_default_10[0 : (n - i - j)]
                if n <= 10 + i + j
                else sc_default_10 + sc_default_26[0 : (n - i - j - 10)]
                if n <= 36 + i + j
                else sc_default_10 + sc_default_26 + sc_default_64[0 : (n - i - j)]
                if n <= 100 + i + j
                else ["grey"] * n
            )
        else:
            sc_default_20 = sc.plotting.palettes.default_20
            sc_default_28 = sc.plotting.palettes.default_28
            sc_default_102 = sc.plotting.palettes.default_102
            palette = (
                sc_default_20[0 : (n - i - j)]
                if n <= 20 + i + j
                else sc_default_20 + sc_default_28[0 : (n - i - j - 20)]
                if n <= 48 + i + j
                else sc_default_102[0 : (n - i - j)]
                if n <= 102 + i + j
                else sc_default_102 + sc_default_102[0 : (n - 102 - i - j)]
                if n <= 204 + i + j
                else ["grey"] * n
            )
    else:
        color_map = plt.get_cmap(cmap)
        palette = [to_hex(color_map(k)) for k in range(n - i - j)]

    if hide_first:
        palette.insert(0, hide_color)
    if hide_last:
        palette.append(hide_color)
    return palette


def clear_colors(ad, slots=None):
    color_slots = [k for k in ad.uns.keys() if k.endswith("_colors")]
    if isinstance(slots, str):
        slots = [slots]
    elif slots is None:
        slots = [s.replace("_colors", "") for s in color_slots]
    if isinstance(slots, (tuple, list)):
        slots = [s for s in slots if f"{s}_colors" in color_slots]
    for s in slots:
        del ad.uns[f"{s}_colors"]


def _is_numeric(x):
    return x.dtype.kind in ("i", "f")


def abline(slope=1, intercept=0, ax=None, log=False, min_x=1):
    """Plot a line from slope and intercept"""
    if ax is None:
        ax = plt.gca()
    x_vals = np.array(ax.get_xlim())
    if log:
        if slope >= 0:
            x_vals[0] = max(min_x, (min_x - intercept) / slope, x_vals[0])
        else:
            x_vals[0] = max(min_x, x_vals[0])
            x_vals[1] = min((min_x - intercept) / slope, x_vals[1])
    y_vals = intercept + slope * x_vals
    ax.plot(x_vals, y_vals, "--", c="k")


def heatmap(
    tbl,
    stylize=None,
    cluster=False,
    figsize=(4, 4),
    **kwargs,
):
    """plot a table as a heatmap"""
    if stylize is not None:
        return tbl.style.background_gradient(cmap="viridis", axis=stylize)

    if cluster:
        fig = sn.clustermap(tbl, linewidths=0.01, figsize=figsize, **kwargs)
    else:
        set_figsize(figsize)
        fig = sn.heatmap(tbl, linewidths=0.01, **kwargs)
    return fig


def set_figsize(dim):
    if len(dim) == 2 and (isinstance(dim[0], (int, float)) and isinstance(dim[1], (int, float))):
        rcParams.update({"figure.figsize": dim})
    else:
        raise ValueError(
            f"Invalid {dim} value, must be an iterable of "
            "length two in the form of (width, height)."
        )


def plot_df_heatmap(
    df,
    cmap="viridis",
    title=None,
    figsize=(7, 7),
    rotation=90,
    save=None,
    **kwargs,
):
    fig, ax = plt.subplots(figsize=figsize)
    im = ax.imshow(df, cmap=cmap, aspect="auto", **kwargs)
    if 0 < rotation < 90:
        horizontalalignment = "right"
    else:
        horizontalalignment = "center"
    plt.xticks(
        range(len(df.columns)),
        df.columns,
        rotation=rotation,
        horizontalalignment=horizontalalignment,
    )
    plt.yticks(range(len(df.index)), df.index)
    if title:
        fig.suptitle(title)
    fig.colorbar(im)
    if save:
        plt.savefig(fname=save, bbox_inches="tight", pad_inches=0.1)


def _log_hist(x, bins=50, min_x=None, max_x=None, ax=None):
    x_copy = x.copy()
    if max_x is None:
        max_x = x[np.isfinite(x)].max()
    x_copy[x > max_x] = max_x
    if min_x is None:
        min_x1 = x[x > 0].min()
        min_x = 10 ** (np.log10(min_x1) - (np.log10(max_x) - np.log10(min_x1)) / (bins - 1))
    x_copy[x < min_x] = min_x

    log_bins = np.logspace(np.log10(min_x), np.log10(max_x), bins)
    if ax:
        ax.hist(x_copy, bins=log_bins)
        ax.set_xscale("log")
    else:
        fig, ax = plt.subplots()
        ax.hist(x_copy, bins=log_bins)
        ax.set_xscale("log")
        return ax


def dotplot2(
    adata,
    keys,
    groupby=None,
    min_group_size=0,
    min_presence=0,
    use_raw=None,
    mean_only_expressed=False,
    second_key_dependent_fraction=False,
    coexpression=False,
    vmin=0,
    vmax=1,
    dot_min=None,
    dot_max=None,
    color_map="Reds",
    swap_axis=False,
    legend_loc="right",
    title="",
    title_loc="top",
    title_size=None,
    omit_xlab=False,
    omit_ylab=False,
    xtickslabels=None,
    ytickslabels=None,
    ax=None,
    save=None,
    save_dpi=80,
    return_data=False,
    **kwargs,
):
    if isinstance(keys, str):
        keys = [keys]
    if not isinstance(keys, (list, tuple)):
        raise ValueError("keys must be a tuple/list of obs keys or var_names")

    if second_key_dependent_fraction or coexpression:
        if len(keys) != 2:
            raise ValueError("Exactly two keys requied if `second_key_dependent_fraction=True`")

    cmap = plt.get_cmap(color_map)
    if use_raw or (use_raw is None and adata.raw):
        ad = adata.raw
    else:
        ad = adata

    if groupby is None:
        groups = pd.Series(np.ones(adata.n_obs)).astype(str).astype("category")
        n_group = 1
    elif groupby in adata.obs.columns:
        groups = adata.obs[groupby]
        if groups.dtype.name != "category":
            groups = groups.astype("category")
    else:
        raise ValueError(f"{groupby} not found")
    groups.cat.remove_unused_categories(inplace=True)
    grouping = list(groups.cat.categories)

    n_obs = ad.shape[0]
    n_key = len(keys)
    y = np.zeros((n_key, n_obs))
    for i in range(n_key):
        key = keys[i]
        if key in adata.obs.columns:
            y[i] = adata.obs[key].values
        elif key in ad.var_names:
            if sc.__version__.startswith("1.4"):
                y[i] = ad[:, key].X
            else:
                y[i] = ad.obs_vector(key)
        else:
            sc_warn(f"{key} not found")
            y[i] = np.zeros(n_obs)

    df = pd.DataFrame(y.T, columns=keys)
    df["group"] = groups.reset_index(drop=True)
    df["group"].cat.set_categories(grouping, inplace=True)
    group_size = df.groupby("group").agg("size")
    if min_group_size:
        df = df.loc[df["group"].isin(group_size.index[group_size.values >= min_group_size]), :]
        df["group"].cat.remove_unused_categories(inplace=True)
    n_group = df["group"].cat.categories.size

    if second_key_dependent_fraction:
        exp_cnt = (
            df.groupby("group")[keys].apply(lambda g: ((g > vmin).sum(axis=1) == 2).sum()).values
        )
        frac = df.groupby("group")[keys].apply(
            lambda g: ((g > vmin).sum(axis=1) == 2).astype(int).mean()
        )
        group_label = frac.index.values
        frac = frac.values
        avg0 = df.groupby("group")[[keys[0]]].apply(lambda g: g.mean(axis=0)).values
        avg1 = (
            df.groupby("group")[[keys[0]]]
            .apply(lambda g: g.sum(axis=0) / (g > vmin).sum(axis=0))
            .fillna(0)
            .values
        )
        avg = avg1 if mean_only_expressed else avg0
        keys = ["|".join(keys)]
        n_key = 1
    elif coexpression:
        exp_cnt = (
            df.groupby("group")[keys].apply(lambda g: ((g > vmin).sum(axis=1) == 2).sum()).values
        )
        frac = df.groupby("group")[keys].apply(
            lambda g: ((g > vmin).sum(axis=1) == 2).astype(int).mean()
        )
        group_label = frac.index.values
        frac = frac.values
        avg0 = df.groupby("group")[keys].apply(lambda g: g.min(axis=1).mean()).values
        avg1 = (
            df.groupby("group")[keys]
            .apply(lambda g: g.min(axis=1).sum() / max(1, (((g > vmin).sum(axis=1) == 2)).sum()))
            .fillna(0)
            .values
        )
        avg = avg1 if mean_only_expressed else avg0
        keys = ["+".join(keys)]
        n_key = 1
    else:
        exp_cnt = df.groupby("group")[keys].apply(lambda g: (g > vmin).sum(axis=0)).values
        frac = df.groupby("group")[keys].apply(lambda g: (g > vmin).mean(axis=0))
        group_label = frac.index.values
        frac = frac.values
        avg0 = df.groupby("group")[keys].apply(lambda g: g.mean(axis=0)).values
        avg1 = (
            df.groupby("group")[keys]
            .apply(lambda g: g.sum(axis=0) / (g > vmin).sum(axis=0))
            .fillna(0)
            .values
        )
        avg = avg1 if mean_only_expressed else avg0

    frac[exp_cnt < min_presence] = 0
    avg[exp_cnt < min_presence] = 0

    data = pd.DataFrame(
        {
            "gene": np.repeat(keys, n_group),
            "group": np.tile(grouping, n_key),
            "frac": frac.T.flatten(),
            "avg": avg.T.flatten(),
        }
    )
    if return_data:
        return data

    if dot_max is None:
        dot_max = np.ceil(np.max(frac) * 10) / 10
    else:
        if dot_max < 0 or dot_max > 1:
            raise ValueError("`dot_max` value has to be between 0 and 1")
    if dot_min is None:
        dot_min = 0
    else:
        if dot_min < 0 or dot_min > 1:
            raise ValueError("`dot_min` value has to be between 0 and 1")

    if dot_min != 0 or dot_max != 1:
        # clip frac between dot_min and  dot_max
        frac = np.clip(frac, dot_min, dot_max)
        old_range = dot_max - dot_min
        # re-scale frac between 0 and 1
        frac = (frac - dot_min) / old_range

    frac = frac.T.flatten()
    avg = avg.T.flatten()

    dot_sizes = (frac * 10) ** 2
    normalize = Normalize(vmin=vmin, vmax=vmax)
    dot_colors = cmap(normalize(avg))

    legend_loc = "none" if ax else legend_loc
    fig_width = 0.4 + (n_key if swap_axis else n_group) * 0.25 + 0.25 * int(legend_loc == "right")
    fig_height = 0.4 + (n_group if swap_axis else n_key) * 0.2 + 0.25 * int(legend_loc == "bottom")

    rcParams.update({"figure.figsize": (fig_width, fig_height)})
    if legend_loc == "right":
        fig, axs = plt.subplots(
            ncols=2,
            nrows=1,
            gridspec_kw={
                "width_ratios": [fig_width - 0.25, 0.25],
                "wspace": 0.25 / (n_key if swap_axis else n_group),
            },
        )
    elif legend_loc == "bottom":
        fig, axs = plt.subplots(
            ncols=1,
            nrows=2,
            gridspec_kw={
                "height_ratios": [fig_height - 0.25, 0.25],
                "hspace": 0.25 / (n_group if swap_axis else n_key),
            },
        )
    elif not ax:
        fig, axs = plt.subplots(ncols=1, nrows=1)
        axs = [axs]
    else:
        axs = [ax]
    main = axs[0]

    if legend_loc == "bottom":
        main.xaxis.tick_top()
    if swap_axis:
        main.scatter(
            y=np.tile(range(n_group), n_key),
            x=np.repeat(np.arange(n_key)[::-1], n_group),
            color=dot_colors[::-1],
            s=dot_sizes[::-1],
            cmap=cmap,
            norm=None,
            edgecolor="none",
            **kwargs,
        )
        main.set_xticks(range(n_key))
        if xtickslabels is None:
            xtickslabels = keys
        main.set_xticklabels(xtickslabels, rotation=270)
        main.set_xlim(-0.5, n_key - 0.5)
        main.set_yticks(range(n_group))
        if ytickslabels is None:
            ytickslabels = group_label[::-1]
        main.set_yticklabels(ytickslabels)
        main.set_ylim(-0.5, n_group - 0.5)
    else:
        main.scatter(
            x=np.tile(range(n_group), n_key),
            y=np.repeat(np.arange(n_key), n_group),
            color=dot_colors,
            s=dot_sizes,
            cmap=cmap,
            norm=None,
            edgecolor="none",
            **kwargs,
        )
        main.set_yticks(range(n_key))
        if ytickslabels is None:
            ytickslabels = keys
        main.set_yticklabels(ytickslabels)
        main.set_ylim(-0.5, n_key - 0.5)
        main.set_xticks(range(n_group))
        if xtickslabels is None:
            xtickslabels = group_label
        main.set_xticklabels(xtickslabels, rotation=270)
        main.set_xlim(-0.5, n_group - 0.5)
    if title:
        if title_loc == "top":
            main.set_title(title, fontsize=15)
        elif title_loc == "right":
            title_ax = axs[1] if legend_loc == "right" else main
            ylab_position = "left" if legend_loc == "right" else "right"
            ylab_pad = 0 if legend_loc == "right" else 20
            title_ax.yaxis.set_label_position(ylab_position)
            if title_size is None:
                title_size = min(15, 100 * fig_height / len(title))
            title_ax.set_ylabel(title, rotation=270, labelpad=ylab_pad, fontsize=title_size)
    if omit_xlab:
        main.tick_params(axis="x", bottom=False, labelbottom=False)
    if omit_ylab:
        main.tick_params(axis="y", left=False, labelleft=False)

    if legend_loc in ("right", "bottom"):
        diff = dot_max - dot_min
        if 0.2 < diff <= 0.6:
            step = 0.1
        elif 0.06 < diff <= 0.2:
            step = 0.05
        elif 0.03 < diff <= 0.06:
            step = 0.02
        elif diff <= 0.03:
            step = 0.01
        else:
            step = 0.2
        # a descending range that is afterwards inverted is used
        # to guarantee that dot_max is in the legend.
        fracs_legends = np.arange(dot_max, dot_min, step * -1)[::-1]
        if dot_min != 0 or dot_max != 1:
            fracs_values = (fracs_legends - dot_min) / old_range
        else:
            fracs_values = fracs_legends
        size = (fracs_values * 10) ** 2
        # color = [
        #    cmap(normalize(value)) for value in np.repeat(vmin + (vmax - vmin) * 0.8, len(size))
        # ]

        # plot size bar
        size_legend = axs[1]
        labels = ["{:.0%}".format(x) for x in fracs_legends]
        if dot_max < 1:
            labels[-1] = ">=" + labels[-1]

        if legend_loc == "bottom":
            size_legend.scatter(y=np.repeat(0, len(size)), x=range(len(size)), s=size, color="grey")
            size_legend.set_xticks(range(len(size)))
            # size_legend.set_xticklabels(labels, rotation=270)
            size_legend.set_xticklabels(["{:.0%}".format(x) for x in fracs_legends], rotation=270)
            # remove y ticks and labels
            size_legend.tick_params(axis="x", bottom=False, labelbottom=True, pad=-2)
            size_legend.tick_params(axis="y", left=False, labelleft=False)
            xmin, xmax = size_legend.get_xlim()
            size_legend.set_xlim(xmin - ((n_key if swap_axis else n_group) - 1) * 0.75, xmax + 0.5)
        else:
            size_legend.scatter(np.repeat(0, len(size)), range(len(size)), s=size, color="grey")
            size_legend.set_yticks(range(len(size)))
            # size_legend.set_yticklabels(labels)
            size_legend.set_yticklabels(["{:.0%}".format(x) for x in fracs_legends])
            # remove x ticks and labels
            size_legend.tick_params(axis="y", left=False, labelleft=False, labelright=True, pad=-2)
            size_legend.tick_params(axis="x", bottom=False, labelbottom=False)
            ymin, ymax = size_legend.get_ylim()
            size_legend.set_ylim(ymin - ((n_group if swap_axis else n_key) - 1) * 0.75, ymax + 0.5)

        # remove surrounding lines
        size_legend.spines["right"].set_visible(False)
        size_legend.spines["top"].set_visible(False)
        size_legend.spines["left"].set_visible(False)
        size_legend.spines["bottom"].set_visible(False)
        size_legend.grid(False)

    if save:
        fig.savefig(save, bbox_inches="tight", dpi=save_dpi)
        plt.close()

    if ax:
        return fig_width, fig_height

    return main


def dotplot_combined_coexpression(
    ad, genes, groupby, groups=None, merge=False, title="", save=None, save_dpi=200, **kwargs
):
    print(save)
    n_gene = len(genes)
    assert n_gene == 2, "genes must be "

    fig, ax = plt.subplots(ncols=n_gene + 1, gridspec_kw={"wspace": 0 if merge else 0.05})
    fW, fH = 0, 0
    if groups is not None and len(groups) != len(ad.obs[groupby].unique()):
        ad = ad[ad.obs[groupby].isin(groups), :]
    for i, gene in enumerate(genes):
        w, h = dotplot2(
            ad,
            gene,
            groupby=groupby,
            swap_axis=True,
            ax=ax[i],
            omit_xlab=False,
            omit_ylab=i > 0,
            title="",
            **kwargs,
        )
        if merge:
            if i < n_gene - 1:
                ax[i].spines["right"].set_visible(False)
            else:
                ax[i].spines["left"].set_visible(False)
        fW += w
        fH = max(fH, h)
    i += 1
    w, h = dotplot2(
        ad,
        genes,
        groupby=groupby,
        coexpression=True,
        ax=ax[i],
        swap_axis=True,
        title=title,
        title_loc="right",
        omit_xlab=False,
        omit_ylab=True,
        **kwargs,
    )
    fW += w
    fH = max(fH, h)

    fig.set_figwidth(fW)
    fig.set_figheight(fH)
    if save:
        fig.savefig(save, bbox_inches="tight", dpi=save_dpi)
        plt.close()


def dotplot3(adata, groupby, use_rep="X", order_gene_by="frac", order_group_by="avg", **kwargs):
    from scipy.spatial.distance import pdist
    from sklearn.preprocessing import minmax_scale
    from seriate import seriate

    if order_gene_by not in ("frac", "avg", None):
        raise ValueError("`order_gene_by` must be one of 'frac', 'avg' or None")
    if order_group_by not in ("frac", "avg", None):
        raise ValueError("`order_group_by` must be one of 'frac', 'avg' or None")

    dotplot_data = summarise_expression_by_group(adata, groupby, use_rep=use_rep)

    if order_gene_by is not None:
        gene_order_df = pd.pivot(dotplot_data, "gene", "group", order_gene_by)
        gene_order = seriate(
            pdist(gene_order_df.values), approximation_multiplier=10000, timeout=10
        )
        ordered_genes = gene_order_df.index[gene_order].tolist()
    else:
        ordered_genes = gene_order_df.index.tolist()

    if order_group_by is not None:
        group_order_df = pd.pivot(dotplot_data, "gene", "group", order_group_by).loc[ordered_genes]
        group_order_df = pd.DataFrame(
            minmax_scale(group_order_df.values),
            index=group_order_df.index,
            columns=group_order_df.columns,
        )
        group_pos = np.zeros(group_order_df.shape[1], dtype=np.int32)
        for i in range(gene_order_df.shape[1]):
            x = group_order_df.iloc[:, i]
            good_indices = np.where(x >= 0.5)[0][0]
            if good_indices.size == 0:
                good_indices = np.where(x == x.max())[0]
            group_pos[i] = good_indices.mean()
        group_order = np.argsort(group_pos)
        ordered_groups = gene_order_df.columns[group_order].tolist()

    old_group_order = adata.obs[groupby].cat.categories.copy()
    try:
        if order_group_by is not None:
            adata.obs[groupby].cat.reorder_categories(ordered_groups, inplace=True)
        sc.pl.dotplot(adata, ordered_genes, groupby=groupby, **kwargs)
    finally:
        adata.obs[groupby].cat.reorder_categories(old_group_order, inplace=True)


def plot_qc_violin(
    adata,
    metrics=None,
    groupby=None,
    one_per_line=False,
    rotation=0,
    figsize=(4, 3),
    return_fig=False,
    **kwargs,
):
    kwargs["linewidth"] = kwargs.get("linewidth", 0.2)
    kwargs["width"] = kwargs.get("width", 1)
    kwargs["scale"] = kwargs.get("scale", "count")
    obs_df = adata.obs.copy()
    if "log1p_n_counts" not in obs_df.columns and "n_counts" in obs_df.columns:
        obs_df["log1p_n_counts"] = np.log1p(obs_df["n_counts"])
    if "log1p_n_genes" not in obs_df.columns and "n_genes" in obs_df.columns:
        obs_df["log1p_n_genes"] = np.log1p(obs_df["n_genes"])
    qc_metrics = ["log1p_n_counts", "log1p_n_genes"] + obs_df.columns[
        obs_df.columns.str.startswith("percent_")
    ].to_list()
    if "scrublet_score_z" in obs_df.columns:
        qc_metrics.append("scrublet_score_z")
    if isinstance(metrics, (tuple, list)):
        qc_metrics = metrics
    for qmt in qc_metrics:
        if qmt not in obs_df.columns:
            raise ValueError(f"{qmt} not found.")
    if groupby is None:
        df0 = obs_df[qc_metrics]
        groupby = "__group__"
        df0[groupby] = "all"
    else:
        df0 = obs_df[[groupby] + qc_metrics]
    df1 = pd.melt(
        df0, id_vars=[groupby], value_vars=qc_metrics, var_name="metric", value_name="value"
    )

    g = sn.catplot(
        data=df1,
        x=groupby,
        y="value",
        row="metric" if one_per_line else None,
        col="metric" if not one_per_line else None,
        sharex=one_per_line,
        sharey=False,
        height=figsize[1],
        aspect=figsize[0] / figsize[1],
        kind="violin",
        inner=None,
        palette=adata.uns.get(
            f"{groupby}_colors", make_palette(adata.obs[groupby].cat.categories.size)
        ),
        **kwargs,
    )
    g.set_xticklabels(rotation=rotation)
    for i, qmt in enumerate(qc_metrics):
        g.facet_axis(i if one_per_line else 0, 0 if one_per_line else i).set_ylabel(qmt)
        if groupby == "__group__":
            g.facet_axis(
                i if one_per_line else 0, 0 if one_per_line else i
            ).axes.get_xaxis().set_visible(False)
        else:
            if one_per_line:
                g.facet_axis(i, 0).set_title("")
                if i < len(qc_metrics) - 1:
                    g.facet_axis(i, 0).axes.get_xaxis().set_visible(False)
    if one_per_line:
        g.fig.subplots_adjust(hspace=0.05)
    if return_fig:
        return g.fig


def plot_qc_scatter(
    adata,
    metric_pairs=None,
    color_by=None,
    show_group_size=False,
    use_hexbin=False,
    figsize=(3, 3),
    wspace=0.25,
    axs=None,
    return_fig=False,
    **kwargs,
):
    obs_df = adata.obs.copy()
    n_cell = obs_df.shape[0]
    if metric_pairs is None:
        if "n_counts" not in obs_df.columns:
            raise ValueError("required metric `n_counts` missing from `obs`")
        metric_pairs = []
        if "n_genes" in obs_df.columns:
            metric_pairs.append(("log1p_n_counts", "log1p_n_genes"))
        for pct_m in (
            "percent_mito",
            "percent_ribo",
            "percent_top50",
            "percent_soup",
            "percent_spliced",
        ):
            if pct_m in obs_df.columns:
                metric_pairs.append(("log1p_n_counts", pct_m))
        if "percent_mito" in obs_df.columns and "percent_ribo" in obs_df.columns:
            metric_pairs.append(("log(percent_mito)", "percent_ribo"))
        if "percent_soup" in obs_df.columns and "percent_spliced" in obs_df.columns:
            metric_pairs.append(("log(percent_soup)", "percent_spliced"))
    n_pair = len(metric_pairs)

    def _logged_metric(m):
        return m.startswith("log1p_") or (m.startswith("log(") and m.endswith(")"))

    def _unlog(m):
        if m.startswith("log1p_"):
            return m[6:]
        if m.startswith("log(") and m.endswith(")"):
            return m[4:-1]
        return m

    scales = [
        (
            "log" if _logged_metric(mp[0]) else "linear",
            "log" if _logged_metric(mp[1]) else "linear",
        )
        for mp in metric_pairs
    ]
    metric_pairs = [(_unlog(mp[0]), _unlog(mp[1])) for mp in metric_pairs]

    old_figsize = rcParams.get("figure.figsize")
    rcParams.update({"figure.figsize": (n_pair * figsize[0], figsize[1])})
    try:
        if axs is None:
            fig, axs = plt.subplots(ncols=n_pair, nrows=1, gridspec_kw={"wspace": wspace})
            if n_pair == 1:
                axs = [axs]
        else:
            fig = None
        point_size = max(kwargs.pop("s", 1 / n_cell), 1e-5)
        kwargs["rasterized"] = kwargs.get("rasterized", True)
        if color_by in obs_df.columns:
            if obs_df[color_by].dtype.kind in ("i", "f"):
                color_var = obs_df[color_by].values
                # vmin = kwargs.pop("vmin", color_var.min())
                # vmax = kwargs.pop("vmax", color_var.max())
                # color_var = (color_var - vmin) / (vmax - vmin)
                for i, mp in enumerate(metric_pairs):
                    m1, m2 = mp
                    panel = axs[i].scatter(
                        x=obs_df[m1],
                        y=obs_df[m2],
                        c=color_var,
                        cmap=kwargs.pop("cmap", "viridis"),
                        s=point_size,
                        **kwargs,
                    )
                    axs[i].set_xlabel(m1)
                    axs[i].set_ylabel(m2)
                    axs[i].set_xscale(scales[i][0])
                    axs[i].set_yscale(scales[i][1])
                if fig is not None:
                    fig.subplots_adjust(right=1 - 0.2 / (i + 1))
                    cbar_ax = fig.add_axes([1 - 0.2 / (i + 1) * 0.75, 0.15, 0.05 / (i + 1), 0.7])
                    fig.colorbar(panel, cax=cbar_ax)
            else:
                if pd.api.types.is_categorical_dtype(obs_df[color_by]):
                    color_var = obs_df[color_by].cat.categories
                else:
                    color_var = obs_df[color_by].astype("category").cat.categories
                n_color = color_var.size
                palette = adata.uns.get(f"{color_by}_colors", make_palette(n_color))
                legend_elements = [
                    Line2D(
                        [0],
                        [0],
                        marker="o",
                        label=f"{v}, n={sum(obs_df[color_by] == v)}" if show_group_size else v,
                        color="white",
                        markerfacecolor=c,
                    )
                    for c, v in zip(palette, color_var)
                ]
                alphas = np.repeat(0.1, n_color)
                alphas[0] = 1
                for i, mp in enumerate(metric_pairs):
                    m1, m2 = mp
                    for j, v in enumerate(color_var):
                        k = obs_df[color_by] == v
                        if use_hexbin:
                            cm = LinearSegmentedColormap.from_list(
                                "custom gradient", ["#FFFFFF", palette[j]], N=256
                            )
                            axs[i].hexbin(
                                x=obs_df.loc[k, m1],
                                y=obs_df.loc[k, m2],
                                mincnt=1,
                                cmap=cm,
                                alpha=alphas[j],
                                **kwargs,
                            )
                        else:
                            axs[i].scatter(
                                x=obs_df.loc[k, m1],
                                y=obs_df.loc[k, m2],
                                color=palette[j],
                                s=point_size,
                                **kwargs,
                            )
                    axs[i].set_xlabel(m1)
                    axs[i].set_ylabel(m2)
                    axs[i].set_xscale(scales[i][0])
                    axs[i].set_yscale(scales[i][1])
                axs[n_pair - 1].legend(
                    handles=legend_elements,
                    loc="center left",
                    bbox_to_anchor=[1, 0, 0, 1],
                    ncol=max(1, len(legend_elements) // 10),
                )
        else:
            for i, mp in enumerate(metric_pairs):
                m1, m2 = mp
                if use_hexbin:
                    axs[i].hexbin(
                        x=obs_df[m1],
                        y=obs_df[m2],
                        mincnt=1,
                        xscale=scales[i][0],
                        yscale=scales[i][1],
                        cmap="viridis",
                    )
                else:
                    axs[i].scatter(x=obs_df[m1], y=obs_df[m2], color="k", s=point_size)
                    axs[i].set_xscale(scales[i][0])
                    axs[i].set_yscale(scales[i][1])
                axs[i].set_xlabel(m1)
                axs[i].set_ylabel(m2)
        if return_fig:
            return fig
    finally:
        rcParams.update({"figure.figsize": old_figsize})


def plot_metric_by_rank(
    adata,
    subject="cell",
    metric="n_counts",
    kind="rank",
    nbins=50,
    order=True,
    decreasing=True,
    ax=None,
    title=None,
    hpos=None,
    vpos=None,
    logx=True,
    logy=True,
    swap_axis=False,
    **kwargs,
):
    """Plot metric by rank from top to bottom"""
    kwargs["c"] = kwargs.get("c", "black")
    metric_names = {
        "n_counts": "nUMI",
        "n_genes": "nGene",
        "n_cells": "nCell",
        "percent_mito": "fracMT",
        "percent_viral": "fracViral",
    }
    if subject not in ("cell", "gene"):
        print('`subject` must be "cell" or "gene".')
        return
    # if metric not in metric_names:
    #    print('`metric` must be "n_counts", "n_genes", "n_cells" or "percent_mito".')
    #    return

    if metric == "percent_mito" and "percent_mito" not in adata.obs.columns:
        if "mito" in adata.var.columns:
            k_mt = adata.var.mito.astype(bool).values
        else:
            k_mt = adata.var_names.str.startswith("MT-")
        if sum(k_mt) > 0:
            adata.obs["percent_mito"] = (
                np.squeeze(np.asarray(adata.X[:, k_mt].sum(axis=1))) / adata.obs["n_counts"] * 100
            )

    if not ax:
        fig, ax = plt.subplots()

    if kind == "rank":
        order_modifier = -1 if decreasing else 1

        if subject == "cell":
            if metric not in adata.obs.columns:
                if metric == "n_counts":
                    adata.obs["n_counts"] = adata.X.sum(axis=1).A1
                if metric == "n_genes":
                    adata.obs["n_genes"] = (adata.X > 0).sum(axis=1).A1
            if swap_axis:
                x = 1 + np.arange(adata.shape[0])
                y = adata.obs[metric].values
                if order is not False:
                    k = np.argsort(order_modifier * y) if order is True else order
                    y = y[k]
            else:
                y = 1 + np.arange(adata.shape[0])
                x = adata.obs[metric].values
                if order is not False:
                    k = np.argsort(order_modifier * x) if order is True else order
                    x = x[k]
        else:
            if metric not in adata.var.columns:
                if metric == "n_counts":
                    adata.var["n_counts"] = adata.X.sum(axis=0).A1
            if swap_axis:
                x = 1 + np.arange(adata.shape[1])
                y = adata.var[metric].values
                if order is not False:
                    k = np.argsort(order_modifier * y) if order is True else order
                    y = y[k]
            else:
                y = 1 + np.arange(adata.shape[1])
                x = adata.var[metric].values
                if order is not False:
                    k = np.argsort(order_modifier * x) if order is True else order
                    x = x[k]

        if kwargs["c"] is not None and not isinstance(kwargs["c"], str):
            kwargs_c = kwargs["c"][k]
            del kwargs["c"]
            ax.scatter(x, y, c=kwargs_c, **kwargs)
        else:
            marker = kwargs.get("marker", "-")
            ax.plot(x, y, marker, **kwargs)

        if logy:
            ax.set_yscale("log")
        if swap_axis:
            ax.set_xlabel("Rank")
        else:
            ax.set_ylabel("Rank")

    elif kind == "hist":
        del kwargs["c"]
        if subject == "cell":
            value = adata.obs[metric]
            logbins = np.logspace(
                np.log10(np.min(value[value > 0])), np.log10(np.max(adata.obs[metric])), nbins
            )
            h = ax.hist(value[value > 0], logbins, **kwargs)
            y = h[0]
            x = h[1]
        else:
            value = adata.var[metric]
            logbins = np.logspace(
                np.log10(np.min(value[value > 0])), np.log10(np.max(adata.var[metric])), nbins
            )
            h = ax.hist(value[value > 0], logbins, **kwargs)
            y = h[0]
            x = h[1]

        ax.set_ylabel("Count")

    if hpos is not None:
        ax.hlines(hpos, xmin=x.min(), xmax=x.max(), linewidth=1, colors="red")
    if vpos is not None:
        ax.vlines(vpos, ymin=y.min(), ymax=y.max(), linewidth=1, colors="green")
    if logx:
        ax.set_xscale("log")
    if swap_axis:
        ax.set_ylabel(metric_names.get(metric, metric))
    else:
        ax.set_xlabel(metric_names.get(metric, metric))
    if title:
        ax.set_title(title)
    ax.grid(linestyle="--")
    if "label" in kwargs:
        ax.legend(loc="upper right" if decreasing else "upper left")

    return ax


def plot_embedding(
    adata,
    groupby,
    basis="umap",
    color=None,
    annot=True,
    min_group_size=0,
    greyout_group=None,
    highlight_group=None,
    size=None,
    use_uns_colors=True,
    save=None,
    savedpi=300,
    figsize=(4, 4),
    **kwargs,
):
    set_figsize(figsize)
    if f"X_{basis}" not in adata.obsm.keys():
        raise KeyError(f'"X_{basis}" not found in `adata.obsm`.')
    if min_group_size:
        group_sizes = adata.obs[groupby].value_counts()
        good_groups = group_sizes[group_sizes >= min_group_size].index
        adata = adata[adata.obs[groupby].isin(good_groups)].copy()
    if isinstance(groupby, (list, tuple)):
        groupby = groupby[0]
    if groupby not in adata.obs.columns:
        raise KeyError(f'"{groupby}" not found in `adata.obs`.')
    if adata.obs[groupby].dtype.name != "category":
        if (
            isinstance(adata.obs[groupby][0], (str, bool, np.bool_))
            and adata.obs[groupby].unique().size < 100
        ):
            adata.obs[groupby] = adata.obs[groupby].astype(str).astype("category")
        else:
            sc_warn(f'"{groupby}" is not categorical.')
            plot_scatter(adata, basis=basis, color=groupby, **kwargs)
    categories = list(adata.obs[groupby].cat.categories)
    rename_dict1 = {
        ct: f"{i:^5d} {ct} (n={(adata.obs[groupby]==ct).sum()})" for i, ct in enumerate(categories)
    }
    restore_dict1 = {
        f"{i:^5d} {ct} (n={(adata.obs[groupby]==ct).sum()})": ct for i, ct in enumerate(categories)
    }
    rename_dict2 = {ct: f"{i:^5d} {ct}" for i, ct in enumerate(categories)}
    restore_dict2 = {f"{i:^5d} {ct}": ct for i, ct in enumerate(categories)}

    marker_size = size
    kwargs["show"] = False
    kwargs["save"] = False
    kwargs["frameon"] = kwargs.get("frameon", None)
    kwargs["legend_loc"] = kwargs.get("legend_loc", "right margin")

    if color is not None and color != groupby:
        use_uns_colors = False
    color = groupby if color is None else color
    offset = 0 if "diffmap" in basis else -1
    xi, yi = 1, 2
    if "components" in kwargs:
        xi, yi = kwargs["components"]
    xi += offset
    yi += offset

    if not (use_uns_colors and f"{color}_colors" in adata.uns):
        adata.uns[f"{color}_colors"] = make_palette(
            adata.obs[color].cat.categories.size, kwargs.get("palette", None)
        )
        if greyout_group is not None:
            if isinstance(greyout_group, str):
                greyout_group = [greyout_group]
            greyout_indices = np.where(
                adata.obs[color].cat.categories.isin(greyout_group)
            )[0]
            for nan_i in greyout_indices:
                adata.uns[f"{color}_colors"][nan_i] = "#E0E0E0"

        if highlight_group is not None:
            if isinstance(highlight_group, str):
                highlight_group = [highlight_group]
            nohl_indices = np.where(
                ~adata.obs[color].cat.categories.isin(highlight_group)
            )[0]
            for nohl_i in nohl_indices:
                adata.uns[f"{color}_colors"][nohl_i] = "#E0E0E0"

    if annot == "full":
        adata.obs[groupby].cat.rename_categories(rename_dict1, inplace=True)
    elif annot in (None, False, "none"):
        kwargs["title"] = ""
        kwargs["legend_loc"] = None
    else:
        adata.obs[groupby].cat.rename_categories(rename_dict2, inplace=True)

    try:
        if "ax" in kwargs:
            ax = kwargs["ax"]
            plot_scatter(adata, basis=basis, color=color, size=marker_size, **kwargs)
        else:
            ax = plot_scatter(adata, basis=basis, color=color, size=marker_size, **kwargs)
    finally:
        if annot == "full":
            adata.obs[groupby].cat.rename_categories(restore_dict1, inplace=True)
        elif annot not in (None, False, "none"):
            adata.obs[groupby].cat.rename_categories(restore_dict2, inplace=True)
    if annot not in (None, False, "none"):
        centroids = pseudo_bulk(adata, groupby, use_rep=f"X_{basis}", FUN=np.median).T
        fontsize = kwargs["legend_fontsize"] if "legend_fontsize" in kwargs else 11
        texts = [
            ax.text(x=row[xi], y=row[yi], s=f"{i:d}", fontsize=fontsize - 1.5, fontweight="bold")
            for i, row in centroids.reset_index(drop=True).iterrows()
            if row[0].astype(str) != "nan"
        ]
        from adjustText import adjust_text

        adjust_text(texts, ax=ax, text_from_points=False, autoalign=False)
        version_scale = 1 if sc.__version__.startswith("1.4") else 1
        ax.legend(
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            frameon=False,
            ncol=int(np.ceil(len(categories) / int(figsize[1] * 3) / 8 * fontsize)),
            fontsize=fontsize * 1.1,
            markerscale=3 * (fontsize / 11) * version_scale,
            handletextpad=-1.9,
            labelspacing=9 / fontsize,
        )
    if save:
        plt.savefig(fname=save, dpi=savedpi, bbox_inches="tight", pad_inches=0.1)
    if "ax" not in kwargs:
        return ax


def highlight(
    adata,
    basis,
    groupby,
    groups=None,
    color=None,
    prefix_dict=None,
    hide_rest=False,
    figsize=(4, 4),
    hide_color="#F0F0F0",
    **kwargs,
):
    set_figsize(figsize)
    old_obs = adata.obs.copy()
    if isinstance(prefix_dict, dict):
        groups = {
            k: list(
                adata.obs[groupby].cat.categories[
                    adata.obs[groupby].cat.categories.str.startswith(v)
                ]
            )
            for k, v in prefix_dict.items()
        }
    if groups is None:
        groups = list(adata.obs[groupby].cat.categories)
    if isinstance(groups, (list, tuple)):
        new_obs = pd.get_dummies(adata.obs[groupby])[groups].astype(int)
        if color is not None:
            if color in adata.obs.columns:
                new_obs = new_obs * adata.obs[color].values.astype(float)[:, np.newaxis]
            elif color in adata.raw.var_names:
                new_obs = new_obs * adata.raw[:, color].X.toarray()
        for i, grp in enumerate(groups):
            if grp in adata.var_names:
                new_grp = "c_" + grp
                new_obs.rename(columns={grp: new_grp}, inplace=True)
                groups[i] = new_grp
        adata.obs = new_obs
        try:
            kwargs["color_map"] = "Reds"
            axs = plot_scatter(
                adata, basis=basis, color=groups, legend_loc=None, show=False, **kwargs
            )
            for i, ax in enumerate(axs):
                ax.tick_params(which="both", bottom=False, top=False, left=False, right=False)
                ax.set_xlabel("")
                ax.set_ylabel("")
                plt.gcf().axes[-(i + 1)].remove()

        finally:
            adata.obs = old_obs
            clear_colors(adata)
    elif isinstance(groups, dict):
        new_obs = adata.obs[[groupby]].copy()
        for grp_name, grp in groups.items():
            new_obs[grp_name] = new_obs[groupby].astype(str)
            new_obs.loc[~new_obs[groupby].isin(grp), grp_name] = "others"
            new_obs[grp_name] = (
                new_obs[grp_name].astype("category").cat.reorder_categories(["others"] + grp)
            )
            adata.uns[f"{grp_name}_colors"] = make_palette(
                len(grp) + 1, kwargs.get("palette", None), hide_first=True, hide_color=hide_color
            )
        adata.obs = new_obs
        try:
            if "palette" in kwargs:
                del kwargs["palette"]
            kwargs["title"] = [f"{groupby}: {grp_name}" for grp_name in groups.keys()]
            plot_scatter(adata, basis=basis, color=list(groups.keys()), **kwargs)
        finally:
            adata.obs = old_obs
            clear_colors(adata)


def plot_composition(adata, composition, groupby, sample):
    """Plot composition as a bar plot

    composition: categorical variable in `obs` of which composition is of interest, e.g. cell type
    groupby    : categorical variable in `obs` of which composition is calculated within
    sample     : categorical variable in `obs` across which composition is averaged
    """
    if not ((composition in adata.obs.columns) and (groupby in adata.obs.columns)):
        raise ValueError
    df = (
        pd.merge(
            adata.obs[[sample, groupby]].drop_duplicates().set_index(sample),
            cross_table(adata, sample, composition, normalise="x"),
            left_index=True,
            right_index=True,
            how="right",
        )
        .groupby(groupby)
        .apply(np.mean)
    ) * 100
    fig, ax = plt.subplots(1, 1)
    cmap = (
        ListedColormap(adata.uns[f"{composition}_colors"])
        if f"{composition}_colors" in adata.uns.keys()
        else None
    )
    df.plot.bar(stacked=True, width=0.8, edgecolor="k", linewidth=0.5, cmap=cmap, ax=ax)
    ax.set_ylim(102, -2)
    ax.set_yticklabels([102, 100, 80, 60, 40, 20, 0])
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    return df, fig, ax


def plot_genes(
    adata,
    genes,
    basis="umap",
    gene_symbols=None,
    title_func=None,
    color_map=None,
    figsize=(2, 2),
    xlim=None,
    ylim=None,
    use_hexbin=False,
    save=None,
    **kwargs,
):
    if color_map is None:
        color_map = expression_colormap(0.01)
    var_df = adata.var if adata.raw is None else adata.raw.var
    var_names = var_df[gene_symbols].values if gene_symbols in var_df.columns else var_df.index
    not_found = [g for g in genes if g not in var_names]
    found = [g for g in genes if g in var_names]
    n = len(found)
    sc_warn(f"{n} genes found")
    sc_warn(f'{",".join(not_found)} not found')
    ncols = kwargs.pop("ncols", None) or int(24 / figsize[0])

    titles = list(map(title_func, found)) if title_func else found

    if use_hexbin:
        nrows = int(np.ceil(n / ncols))
        ncols = min(ncols, n)
        set_figsize((figsize[0] * ncols, figsize[1] * nrows))
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, sharey=True, squeeze=False)
        cm = LinearSegmentedColormap.from_list("custom gradient", ["#EEEEEE", "#FF0000"], N=256)
        for i in range(n):
            ax = axs[i // ncols, i % ncols]
            ax.hexbin(
                x=adata.obsm[f"X_{basis}"][:, 0],
                y=adata.obsm[f"X_{basis}"][:, 1],
                C=adata.obs_vector(found[i]),
                cmap=cm,
                **kwargs,
            )
            ax.set_title(titles[i])
            ax.set_xticks([])
            ax.set_yticks([])
        if save:
            fig.savefig(fname=save, bbox_inches="tight", pad_inches=0.1)
    else:
        set_figsize(figsize)
        kwargs["title"] = titles
        axs = plot_scatter(
            adata,
            basis=basis,
            color=found,
            color_map=color_map,
            ncols=ncols,
            wspace=0,
            hspace=0.2,
            show=False,
            gene_symbols=gene_symbols,
            **kwargs,
        )
        if n < 2:
            axs = [axs]
        for i, ax in enumerate(axs):
            ax.tick_params(which="both", bottom=False, top=False, left=False, right=False)
            ax.set_xlabel("")
            ax.set_ylabel("")
            if xlim:
                ax.set_xlim(xlim)
            if ylim:
                ax.set_ylim(xlim)
            plt.gcf().axes[-(i + 1)].remove()
        if save:
            plt.savefig(fname=save, bbox_inches="tight", pad_inches=0.1)


def plot_markers(
    adata: anndata.AnnData,
    groupby: str,
    mks: pd.DataFrame,
    n_genes: int = 5,
    kind: str = "dotplot",
    remove_genes: list = [],
    **kwargs,
):
    df = mks.reset_index()[["index", "top_frac_group"]].rename(
        columns={"index": "gene", "top_frac_group": "cluster"}
    )
    var_tb = adata.raw.var if kwargs.get("use_raw", None) is True or adata.raw else adata.var
    remove_gene_set = set()
    for g_cat in remove_genes:
        if g_cat in var_tb.columns:
            remove_gene_set |= set(var_tb.index[var_tb[g_cat].values])
    df = df[~df.gene.isin(list(remove_gene_set))].copy()
    df1 = df.groupby("cluster").head(n_genes)
    mks_dict = defaultdict(list)
    for c, g in zip(df1.cluster, df1.gene):
        mks_dict[c].append(g)
    func = getattr(sc.pl, kind)
    if sc.__version__.startswith("1.4"):
        return func(adata, df1.gene.to_list(), groupby=groupby, **kwargs)
    else:
        return func(adata, mks_dict, groupby=groupby, **kwargs)


def plot_diffexp(
    adata,
    basis="umap",
    key="rank_genes_groups",
    top_n=4,
    extra_genes=None,
    figsize1=(4, 4),
    figsize2=(2.5, 2.5),
    dotsize=None,
    dotplot=True,
    **kwargs,
):
    grouping = adata.uns[key]["params"]["groupby"]
    de_tbl = extract_de_table(adata.uns[key])
    de_tbl = de_tbl.loc[de_tbl.genes.astype(str) != "nan", :]
    de_genes = list(de_tbl.groupby("cluster").head(top_n)["genes"].values)
    de_clusters = list(de_tbl.groupby("cluster").head(top_n)["cluster"].astype(str).values)
    if extra_genes:
        de_genes.extend(extra_genes)
        de_clusters.extend(["known"] * len(extra_genes))

    rcParams.update({"figure.figsize": figsize1})
    # sc.pl.rank_genes_groups(adata, key=key, show=False)

    if dotplot:
        sc.pl.dotplot(adata, var_names=de_genes, groupby=grouping, show=False)

    expr_cmap = expression_colormap(0.01)
    rcParams.update({"figure.figsize": figsize2})
    plot_scatter(
        adata,
        basis=basis,
        color=de_genes,
        color_map=expr_cmap,
        use_raw=True,
        size=dotsize,
        title=[f"{c}, {g}" for c, g in zip(de_clusters, de_genes)],
        show=False,
        **kwargs,
    )

    rcParams.update({"figure.figsize": figsize1})
    plot_embedding(adata, basis=basis, groupby=grouping, size=dotsize, show=False)


def plot_cellbender_qc(
    ad, raw_nUMI_name="n_counts_raw", nUMI_name="n_counts", title=None, ax=None, overlay_soup=False
):
    if raw_nUMI_name not in ad.obs.columns:
        raise ValueError("Cannot find raw nUMI info in `obs`")

    n_ax = 1
    train_history_found = (
        "test_epoch" in ad.uns.keys()
        and "test_elbo" in ad.uns.keys()
        and "training_elbo_per_epoch" in ad.uns.keys()
    )
    latent_gene_encoding_found = "X_latent_gene_encoding" in ad.obsm.keys()
    if train_history_found:
        n_ax += 1
    if latent_gene_encoding_found:
        n_ax += 1

    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=n_ax, gridspec_kw={"wspace": 0.5})

    raw_nUMI = ad.obs[raw_nUMI_name].values
    if nUMI_name not in ad.obs.columns:
        nUMI = ad.X.sum(axis=1).A1
    else:
        nUMI = ad.obs[nUMI_name].values
    k1 = np.argsort(raw_nUMI)[::-1]
    y1 = raw_nUMI[k1]
    y2 = nUMI[k1]
    cell_prob = ad.obs["latent_cell_probability"].values
    y3 = cell_prob[k1]
    y4 = (1 - nUMI / raw_nUMI)[k1]
    x1 = np.arange(ad.n_obs) + 1
    ax[0].plot(y1 + 1, c="black", label="raw")
    ax[0].scatter(x1, y2 + 1, s=0.1, c="blue", label="CB", rasterized=True)
    ax[0].set_yscale("log")
    ax[0].set_xlabel("Rank")
    ax[0].legend()
    ax[0].set_ylabel("nUMI")
    ax0b = ax[0].twinx()
    ax0b.scatter(x1, y3, s=0.1, c="red", alpha=0.5, label="p", rasterized=True)
    ax0b.set_ylim(-0.02, 1.02)
    if overlay_soup:
        ax0c = ax[0].twinx()
        ax0c.scatter(x1, y4, s=0.1, c="yellow", alpha=0.5, label="f", rasterized=True)
        ax0c.set_ylim(-0.02, 1.02)
        ax0b.set_ylabel("Cell probability (red) / soup fraction (yellow)")
    else:
        ax0b.set_ylabel("Cell probability", color="red")

    if train_history_found:
        y5 = ad.uns["training_elbo_per_epoch"]
        x2 = ad.uns["test_epoch"]
        y6 = ad.uns["test_elbo"]
        ax[1].plot(y5, c="black", label="train")
        ax[1].scatter(x2, y6, s=2, c="blue", label="test", rasterized=True)
        ax[1].set_xlabel("Epoch")
        ax[1].set_ylabel("Elbo")
        ax[1].legend()

    if latent_gene_encoding_found:
        from sklearn.decomposition import PCA

        pca = PCA(n_components=2, svd_solver="arpack")
        pcs = pca.fit_transform(ad.obsm["X_latent_gene_encoding"][cell_prob >= 0.5])
        ax[-1].hexbin(pcs[:, 0], pcs[:, 1], bins="log", cmap="viridis")
        ax[-1].set_xlabel("PC1")
        ax[-1].set_ylabel("PC2")

    if title:
        ax[0].get_figure().suptitle(title, weight="bold")

    if ax is None:
        return fig



"""
scanpy diffexp
"""

import numpy as np
import pandas as pd
import scanpy as sc
from ._utils import sc_warn


def diffexp(
    adata,
    use_raw=True,
    n_genes=None,
    key_added=None,
    logreg_param=None,
    filter_params=None,
    save=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.rank_genes_groups.
    """
    parameters = ["min_in_group_fraction", "max_out_group_fraction", "min_fold_change"]
    if isinstance(filter_params, (tuple, list)) and len(filter_params) == 3:
        filter_params = {parameters[i]: filter_params[i] for i in range(3)}
    elif isinstance(filter_params, dict) and np.all(
        pd.Series(list(filter_params.keys())).isin(parameters)
    ):
        pass
    else:
        if filter_params is not None:
            sc_warn("Unsupported data format for `filter_params`, reset to None")
        filter_params = None

    if adata.raw is None:
        use_raw = False

    if n_genes is None:
        n_genes = adata.raw.shape[1] if use_raw else adata.shape[1]

    if logreg_param and isinstance(logreg_param, dict):
        for key, val in logreg_param.items():
            kwargs[key] = val

    key_added = f"rank_genes_groups_{key_added}" if key_added else "rank_genes_groups"

    sc.tl.rank_genes_groups(adata, use_raw=use_raw, n_genes=n_genes, key_added=key_added, **kwargs)

    de_tbl = extract_de_table(adata.uns[key_added])

    if isinstance(filter_params, dict):
        sc.tl.filter_rank_genes_groups(
            adata,
            key=key_added,
            key_added=key_added + "_filtered",
            use_raw=use_raw,
            **filter_params,
        )
        de_tbl = extract_de_table(adata.uns[key_added + "_filtered"])

    if save:
        de_tbl.to_csv(save, sep="\t", header=True, index=False)

    return de_tbl


def diffexp_paired(adata, groupby, pair, **kwargs):
    """
    Restrict DE to between a pair of clusters, return both up and down genes
    """
    test, ref = pair
    de_key = f"de.{test}-{ref}"
    up_de = diffexp(
        adata,
        key_added=de_key,
        groupby=groupby,
        groups=[test],
        reference=ref,
        **kwargs,
    )
    ref, test = pair
    de_key = f"de.{test}-{ref}"
    down_de = diffexp(
        adata,
        key_added=de_key,
        groupby=groupby,
        groups=[test],
        reference=ref,
        **kwargs,
    )
    return up_de, down_de


def extract_de_table(de_dict):
    """
    Extract DE table from adata.uns
    """
    if de_dict["params"]["method"] == "logreg":
        requested_fields = ("scores",)
    else:
        requested_fields = (
            "scores",
            "logfoldchanges",
            "pvals",
            "pvals_adj",
        )
    gene_df = _recarray_to_dataframe(de_dict["names"], "genes")[["cluster", "rank", "genes"]]
    gene_df["ref"] = de_dict["params"]["reference"]
    gene_df = gene_df[["cluster", "ref", "rank", "genes"]]
    de_df = pd.DataFrame(
        {
            field: _recarray_to_dataframe(de_dict[field], field)[field]
            for field in requested_fields
            if field in de_dict
        }
    )
    de_tbl = gene_df.merge(de_df, left_index=True, right_index=True)
    de_tbl = de_tbl.loc[de_tbl.genes.astype(str) != "nan", :]
    return de_tbl


def _recarray_to_dataframe(array, field_name):
    return (
        pd.DataFrame(array)
        .reset_index()
        .rename(columns={"index": "rank"})
        .melt(id_vars="rank", var_name="cluster", value_name=field_name)
    )
